def ci_label(alt="two-sided", conf=0.95, dec=3):
    if alt == "less":
        return ["0%", f"{round(100 * conf, dec)}%"]
    elif alt == "greater":
        return [f"{round(100 * (1 - conf), dec)}%", "100%"]
    elif alt == "two-sided":
        val = 100 * (1 - conf) / 2
        return [f"{round(v, dec)}%" for v in [val, 100 - val]]
    else:
        return "Error: alt_hyp must be one of 'less', 'greater', or 'two-sided'"



from starlette.applications import Starlette
from starlette.routing import Mount
from starlette.staticfiles import StaticFiles
from starlette.requests import Request as StarletteRequest
from shiny import App, ui, reactive, Inputs, Outputs, Session
from pathlib import Path
import webbrowser
import nest_asyncio
import uvicorn
import os
import signal
import sys
import tempfile
import pyrsm as rsm
import pyrsm.radiant.utils as ru
import pyrsm.radiant.model.model_utils as mu

choices = {
    "None": "None",
    "dist": "Distribution",
    "corr": "Correlation",
    "scatter": "Scatter",
    "dashboard": "Dashboard",
    "residual": "Residual vs Explanatory",
    "pred": "Prediction plots",
    "vimp": "Permutation importance",
    "coef": "Coefficient plot",
}

controls = {
    "ci": "Confidence intervals",
    "rmse": "RMSE",
    "ssq": "Sum of Squares",
    "vif": "VIF",
}


def summary_extra(self):
    return (
        ui.input_radio_buttons(
            "show_interactions",
            "Interactions:",
            selected=self.state.get("show_interactions", None),
            choices={0: "None", 2: "2-way", 3: "3-way"},
            inline=True,
        ),
        ui.panel_conditional(
            "input.show_interactions > 0",
            ui.output_ui("ui_interactions"),
        ),
        ui.input_checkbox_group(
            "controls",
            "Additional output:",
            selected=self.state.get("controls", None),
            choices=controls,
        ),
        ui.output_ui("ui_evar_test"),
    )


def plots_extra(self):
    return (
        ui.panel_conditional(
            "input.plots == 'corr'",
            ui.input_select(
                "nobs",
                "Number of data points plotted:",
                selected=self.state.get("nobs", None),
                choices={1000: "1,000", -1: "All"},
            ),
        ),
        ui.panel_conditional(
            "input.plots == 'pred'",
            ui.output_ui("ui_incl_evar"),
            ui.output_ui("ui_incl_interactions"),
        ),
    )


class model_regress:
    def __init__(
        self, datasets: dict, descriptions=None, state=None, code=True, navbar=None
    ) -> None:
        ru.init(
            self,
            datasets,
            descriptions=descriptions,
            state=state,
            code=code,
            navbar=navbar,
        )

    def shiny_ui(self, request: StarletteRequest):
        return ui.page_navbar(
            ru.head_content(),
            ui.nav_panel(
                "Model > Linear regression (OLS)",
                ui.row(
                    ui.column(
                        3,
                        ru.ui_data(self),
                        ru.ui_summary(summary_extra(self)),
                        mu.ui_predict(self),
                        ru.ui_plot(self, choices, plots_extra(self)),
                    ),
                    ui.column(8, ru.ui_main_model()),
                ),
            ),
            self.navbar,
            ru.ui_help(
                "https://github.com/vnijs/pyrsm/blob/main/examples/model/model-linear-regression.ipynb",
                "Linear regression (OLS) example notebook",
            ),
            ru.ui_stop(),
            title="Radiant for Python",
            inverse=False,
            id="navbar_id",
        )

    def shiny_server(self, input: Inputs, output: Outputs, session: Session):
        # --- section standard for all apps ---
        get_data = ru.make_data_elements(self, input, output, session)

        def update_state():
            with reactive.isolate():
                ru.dct_update(self, input)

        session.on_ended(update_state)

        # --- section standard for all model apps ---
        ru.reestimate(input)

        # --- section unique to each app ---
        mu.make_model_inputs(self, input, output, get_data, "isNum")
        mu.make_int_inputs(self, input, output, get_data)
        show_code, estimate = mu.make_estimate(
            self, input, output, get_data, fun="regress", ret="reg", debug=False
        )
        mu.make_summary(
            self,
            input,
            output,
            session,
            show_code,
            estimate,
            ret="reg",
            sum_fun=rsm.model.regress.summary,
        )
        mu.make_predict(
            self,
            input,
            output,
            session,
            show_code,
            estimate,
            ret="reg",
            pred_fun=rsm.model.regress.predict,
        )
        mu.make_plot(
            self,
            input,
            output,
            session,
            show_code,
            estimate,
            ret="reg",
        )

        # --- section standard for all apps ---
        # stops returning code if moved to utils
        @reactive.Effect
        @reactive.event(input.stop, ignore_none=True)
        async def stop_app():
            rsm.md(f"```python\n{self.stop_code}\n```")
            await session.app.stop()
            os.kill(os.getpid(), signal.SIGTERM)


def regress(
    data_dct: dict = None,
    descriptions_dct: dict = None,
    state: dict = None,
    code: bool = True,
    host: str = "",
    port: int = 8000,
    log_level: str = "warning",
    debug: bool = False,
):
    """
    Launch a Radiant-for-Python app for linear regression analysis
    """
    host = ru.set_host(host)
    if data_dct is None:
        data_dct, descriptions_dct = ru.get_dfs(pkg="model")
    rc = model_regress(data_dct, descriptions_dct, state=state, code=code)
    nest_asyncio.apply()
    webbrowser.open(f"http://{host}:{port}")
    print(f"Listening on http://{host}:{port}")
    ru.message()

    # redirect stdout and stderr to the temporary file
    if not debug:
        org_stdout = sys.stdout
        org_stderr = sys.stderr
        temp = tempfile.NamedTemporaryFile()
        temp_file = open(temp.name, "w")
        sys.stdout = temp_file
        sys.stderr = temp_file

    app = App(rc.shiny_ui, rc.shiny_server)
    www_dir = Path(__file__).parent.parent.parent / "radiant" / "www"
    app_static = StaticFiles(directory=www_dir, html=False)

    routes = [
        Mount("/www", app=app_static),
        Mount("/", app=app),
    ]

    uvicorn.run(
        Starlette(debug=debug, routes=routes),
        host=host,
        port=port,
        log_level=log_level,
    )

    if not debug:
        sys.stdout = org_stdout
        sys.stderr = org_stderr
        temp_file.close()


if __name__ == "__main__":
    regress(debug=False)



import pandas as pd
import numpy as np
from pyrsm.utils import (
    add_description,
    ifelse,
    levels_list,
    expand_grid,
    table2data,
    setdiff,
    union,
    intersect,
)


md = """# Data Description

The variables in the dataset are a and b

* a: The first variable
* b: The second variable"""


def test_add_description():
    df = pd.DataFrame({"col1": [1, 2], "col2": [3, 4]})
    df = add_description(df, md)
    assert df.description == md, "Description not attached"


def test_ifelse_true():
    assert (
        ifelse(3 > 2, "greater", "smaller") == "greater"
    ), "Logical comparison in ifelse incorrect"


def test_ifelse_false():
    assert (
        ifelse(2 > 3, "greater", "smaller") == "smaller"
    ), "Logical comparison in ifelse incorrect"


def test_ifelse_array():
    assert all(
        ifelse(np.array([2, 3, 4]) > 2, 1, 0) == np.array([0, 1, 1])
    ), "Logical comparison of np.array in ifelse incorrect"


df = pd.DataFrame({"var1": ["a", "b", "a"], "var2": [1, 2, 1]})
dct = {"var1": ["a", "b"], "var2": [1, 2]}


def test_level_list():
    assert levels_list(df) == dct, "Levels list created incorrect dictionary"


def test_expand_grid():
    edf = expand_grid(dct)
    assert list(edf.loc[1].values) == ["a", 2], "Expand grid row 1 incorrect"
    assert list(edf.loc[2].values) == ["b", 1], "Expand grid row 3 incorrect"


def test_table2data():
    t2d = table2data(df.assign(freq=[3, 4, 5]), "freq")
    assert t2d.size == 36, "Number of rows from table2data is incorrect"
    assert (
        t2d["var1"] == "a"
    ).sum() == 8, "Number of 'a' values incorrect in table2data"


def test_setdiff():
    assert setdiff(["a", "b", "c"], ["b", "x"]) == [
        "a",
        "c",
    ], "Set difference incorrect"


def test_union():
    assert union(["a", "b", "c"], ["b", "x"]) == ["a", "b", "c", "x"], "Union incorrect"


def test_intersect():
    assert intersect(["a", "b", "c"], ["b", "x"]) == ["b"], "Union incorrect"


if __name__ == "__main__":
    test_add_description()
    test_ifelse_true()
    test_ifelse_false()
    test_ifelse_array()
    test_level_list()
    test_expand_grid()
    test_table2data()
    test_setdiff()
    test_union()
    test_intersect()



import pandas as pd
import os


def convert(base_dir):
    for dirpath, dirnames, filenames in os.walk(base_dir):
        for f in filenames:
            if f.endswith(".pkl"):
                print(f"Converting {f} to parquet")
                df = pd.read_pickle(os.path.join(dirpath, f))
                df.to_parquet(os.path.join(dirpath, f.replace(".pkl", ".parquet")))
                if hasattr(df, "description"):
                    fmd = f.replace(".pkl", "_description.md")
                    md_file = os.path.join(dirpath, fmd)
                    with open(md_file, "w") as file:
                        print(f"Storing description in {fmd}")
                        file.write(df.description)


base_path = os.path.expanduser("~/gh/pyrsm/pyrsm/data")
if os.path.exists(base_path):
    convert(base_path)



from pyrsm.basics.probability_calculator_functions import *

pc_dist = {
    "Binomial": "binom",
    "Chi-squared": "chisq",
    "Discrete": "disc",
    "Exponential": "expo",
    "F": "fdist",
    "Log normal": "lnorm",
    "Normal": "norm",
    "Poisson": "pois",
    "t": "tdist",
    "Uniform": "unif",
}


class prob_calc:
    # Probability calculator
    def __init__(self, distribution: str, **kwargs) -> None:
        self.distribution = distribution
        self.args = kwargs
        if self.distribution == "binom":
            self.dct = prob_binom(**kwargs)
        elif self.distribution == "chisq":
            self.dct = prob_chisq(**kwargs)
        elif self.distribution == "disc":
            self.dct = prob_disc(**kwargs)
        elif self.distribution == "expo":
            self.dct = prob_expo(**kwargs)
        elif self.distribution == "fdist":
            self.dct = prob_fdist(**kwargs)
        elif self.distribution == "lnorm":
            self.dct = prob_lnorm(**kwargs)
        elif self.distribution == "norm":
            self.dct = prob_norm(**kwargs)
        elif self.distribution == "pois":
            self.dct = prob_pois(**kwargs)
        elif self.distribution == "tdist":
            self.dct = prob_tdist(**kwargs)
        elif self.distribution == "unif":
            self.dct = prob_unif(**kwargs)
        else:
            raise ValueError(f"Distribution must be one of {list(pc_dist.keys())}")

    def summary(self, dec=3):
        type = "probs" if ("plb" in self.args) or ("pub" in self.args) else "values"
        if self.distribution == "binom":
            summary_prob_binom(self.dct, type=type, dec=dec)
        elif self.distribution == "chisq":
            summary_prob_chisq(self.dct, type=type, dec=dec)
        elif self.distribution == "disc":
            summary_prob_disc(self.dct, type=type, dec=dec)
        elif self.distribution == "expo":
            summary_prob_expo(self.dct, type=type, dec=dec)
        elif self.distribution == "fdist":
            summary_prob_fdist(self.dct, type=type, dec=dec)
        elif self.distribution == "lnorm":
            summary_prob_lnorm(self.dct, type=type, dec=dec)
        elif self.distribution == "norm":
            summary_prob_norm(self.dct, type=type, dec=dec)
        elif self.distribution == "pois":
            summary_prob_pois(self.dct, type=type, dec=dec)
        elif self.distribution == "tdist":
            summary_prob_tdist(self.dct, type=type, dec=dec)
        elif self.distribution == "unif":
            summary_prob_unif(self.dct, type=type, dec=dec)

    def plot(self):
        type = "probs" if ("plb" in self.args) or ("pub" in self.args) else "values"
        if self.distribution == "binom":
            plot_prob_binom(self.dct, type=type)
        elif self.distribution == "chisq":
            plot_prob_chisq(self.dct, type=type)
        elif self.distribution == "disc":
            plot_prob_disc(self.dct, type=type)
        elif self.distribution == "expo":
            plot_prob_expo(self.dct, type=type)
        elif self.distribution == "fdist":
            plot_prob_fdist(self.dct, type=type)
        elif self.distribution == "lnorm":
            plot_prob_lnorm(self.dct, type=type)
        elif self.distribution == "norm":
            plot_prob_norm(self.dct, type=type)
        elif self.distribution == "pois":
            plot_prob_pois(self.dct, type=type)
        elif self.distribution == "tdist":
            plot_prob_tdist(self.dct, type=type)
        elif self.distribution == "unif":
            plot_prob_unif(self.dct, type=type)


if __name__ == "__main__":
    pc = prob_calc("binom", n=10, p=0.3, lb=1, ub=3)
    pc.summary()
    pc.plot()
    pc = prob_calc("binom", n=10, p=0.1168, lb=4)
    pc.summary()
    pc.plot()
    pc = prob_calc("binom", n=10, p=0.1168, ub=2)
    pc.summary()
    pc.plot()
    pc = prob_calc("binom", n=10, p=0.1168, plb=0.3, pub=0.9)
    pc.summary()
    pc.plot()
    pc = prob_calc("binom", n=10, p=0.1168, pub=0.9)
    pc.summary()
    pc.plot()
    pc = prob_calc("binom", n=10, p=0.1168, plb=0.3)
    pc.summary()
    pc.plot()

    # pc = prob_binom(n=10, p=0.3, lb=1, ub=3)
    # summary_prob_binom(pb, type="values")
    # plot_prob_binom(pb, type="values")
    # pb = prob_binom(n=10, p=0.3, lb=1)
    # summary_prob_binom(pb, type="values")
    # plot_prob_binom(pb, type="values")
    # pb = prob_binom(n=10, p=0.3, ub=3)
    # summary_prob_binom(pb, type="values")
    # plot_prob_binom(pb, type="values")

    # pb = prob_binom(n=10, p=0.3, plb=0.1, pub=0.8)
    # summary_prob_binom(pb, type="probs")
    # plot_prob_binom(pb, type="probs")
    # pb = prob_binom(n=10, p=0.3, plb=0.1)
    # summary_prob_binom(pb, type="probs")
    # plot_prob_binom(pb, type="probs")
    # pb = prob_binom(n=10, p=0.3, pub=0.8)
    # summary_prob_binom(pb, type="probs")
    # plot_prob_binom(pb, type="probs")

    # v = list(range(1, 7))
    # p = [1 / 6] * 6
    # pd = prob_disc(v, p, lb=2, ub=5)
    # summary_prob_disc(pd, type="values")
    # plot_prob_disc(pd, type="values")
    # pd = prob_disc(v, p, lb=2)
    # summary_prob_disc(pd, type="values")
    # plot_prob_disc(pd, type="values")
    # pd = prob_disc(v, p, ub=5)
    # summary_prob_disc(pd, type="values")
    # plot_prob_disc(pd, type="values")

    # v = list(range(1, 7))
    # p = [2 / 6, 2 / 6, 1 / 12, 1 / 12, 1 / 12, 1 / 12]
    # pd = prob_disc(v, p, plb=0.5, pub=0.8)
    # summary_prob_disc(pd, type="probs")
    # plot_prob_disc(pd, type="probs")
    # pd = prob_disc(v, p, plb=0.5)
    # summary_prob_disc(pd, type="probs")
    # plot_prob_disc(pd, type="probs")

    # v = list(range(1, 7))
    # p = [2 / 6, 2 / 6, 1 / 12, 1 / 12, 1 / 12, 1 / 12]
    # pd = prob_disc(v, p, plb=0.05, pub=0.95)
    # summary_prob_disc(pd, type="probs")
    # plot_prob_disc(pd, type="probs")

    # pd = prob_disc(v, p, pub=0.95)
    # summary_prob_disc(pd, type="probs")
    # plot_prob_disc(pd, type="probs")

    # pd = prob_disc(v, p, plb=0.05)
    # summary_prob_disc(pd, type="probs")
    # plot_prob_disc(pd, type="probs")

    # pf = prob_fdist(df1=10, df2=10, lb=0.5, ub=2.978)
    # summary_prob_fdist(pf, type="values")
    # plot_prob_fdist(pf, type="values")
    # pf = prob_fdist(df1=10, df2=10, lb=0.5)
    # summary_prob_fdist(pf, type="values")
    # plot_prob_fdist(pf, type="values")
    # pf = prob_fdist(df1=10, df2=10, ub=2.978)
    # summary_prob_fdist(pf, type="values")
    # plot_prob_fdist(pf, type="values")

    # pf = prob_fdist(df1=10, df2=10, plb=0.05, pub=0.95)
    # summary_prob_fdist(pf, type="probs")
    # plot_prob_fdist(pf, type="probs")
    # pf = prob_fdist(df1=10, df2=10, plb=0.05)
    # summary_prob_fdist(pf, type="probs")
    # pf = prob_fdist(df1=10, df2=10, pub=0.95)
    # summary_prob_fdist(pf, type="probs")
    # plot_prob_fdist(pf, type="probs")

    # pn = prob_norm(mean=0, stdev=1, lb=-0.5, ub=0.5)
    # summary_prob_norm(pn, type="values")
    # plot_prob_norm(pn, type="values")
    # pn = prob_norm(mean=0, stdev=1, lb=-0.5)
    # summary_prob_norm(pn, type="values")
    # plot_prob_norm(pn, type="values")
    # pn = prob_norm(mean=0, stdev=1, ub=0.5)
    # summary_prob_norm(pn, type="values")
    # plot_prob_norm(pn, type="values")

    # pn = prob_norm(mean=0, stdev=1, plb=0.025, pub=0.975)
    # summary_prob_norm(pn, type="probs")
    # plot_prob_norm(pn, type="probs")
    # pn = prob_norm(mean=0, stdev=1, plb=0.025)
    # summary_prob_norm(pn, type="probs")
    # plot_prob_norm(pn, type="probs")
    # pn = prob_norm(mean=0, stdev=1, pub=0.975)
    # summary_prob_norm(pn, type="probs")
    # plot_prob_norm(pn, type="probs")

    # pc = prob_chisq(df=1, lb=1, ub=3.841)
    # summary_prob_chisq(pc, type="values")
    # plot_prob_chisq(pc, type="values")
    # pc = prob_chisq(df=1, lb=1)
    # summary_prob_chisq(pc, type="values")
    # plot_prob_chisq(pc, type="values")
    # pc = prob_chisq(df=1, ub=3.841)
    # summary_prob_chisq(pc, type="values")
    # plot_prob_chisq(pc, type="values")

    # pc = prob_chisq(df=1, plb=0.05, pub=0.95)
    # summary_prob_chisq(pc, type="probs")
    # plot_prob_chisq(pc, type="probs")
    # pc = prob_chisq(df=1, plb=0.05)
    # summary_prob_chisq(pc, type="probs")
    # plot_prob_chisq(pc, type="probs")
    # pc = prob_chisq(df=1, pub=0.95)
    # summary_prob_chisq(pc, type="probs")
    # plot_prob_chisq(pc, type="probs")

    # pu = prob_unif(min=0, max=1, lb=0.2, ub=0.8)
    # summary_prob_unif(pu, type="values")
    # plot_prob_unif(pu, type="values")
    # pu = prob_unif(min=0, max=1, lb=0.2)
    # summary_prob_unif(pu, type="values")
    # plot_prob_unif(pu, type="values")
    # pu = prob_unif(min=0, max=1, ub=0.8)
    # summary_prob_unif(pu, type="values")
    # plot_prob_unif(pu, type="values")

    # pu = prob_unif(min=0, max=1, plb=0.2, pub=0.8)
    # summary_prob_unif(pu, type="probs")
    # plot_prob_unif(pu, type="probs")
    # pu = prob_unif(min=0, max=1, plb=0.2)
    # summary_prob_unif(pu, type="probs")
    # plot_prob_unif(pu, type="probs")
    # pu = prob_unif(min=0, max=1, pub=0.8)
    # summary_prob_unif(pu, type="probs")
    # plot_prob_unif(pu, type="probs")

    # pt = prob_tdist(df=10, lb=-2.228, ub=2.228)
    # summary_prob_tdist(pt, type="values")
    # plot_prob_tdist(pt, type="values")
    # pt = prob_tdist(df=10, ub=2.228)
    # summary_prob_tdist(pt, type="values")
    # plot_prob_tdist(pt, type="values")
    # pt = prob_tdist(df=10, lb=-2.228)
    # plot_prob_tdist(pt, type="values")
    # pt = prob_tdist(df=10)
    # summary_prob_tdist(pt, type="values")
    # plot_prob_tdist(pt, type="values")

    # pt = prob_tdist(df=10, plb=0.025, pub=0.975)
    # summary_prob_tdist(pt, type="probs")
    # plot_prob_tdist(pt, type="prob")
    # pt = prob_tdist(df=10, pub=0.975)
    # summary_prob_tdist(pt, type="probs")
    # plot_prob_tdist(pt, type="probs")
    # pt = prob_tdist(df=10, plb=0.025)
    # summary_prob_tdist(pt, type="probs")
    # plot_prob_tdist(pt, type="probs")



from starlette.applications import Starlette
from starlette.routing import Mount
from starlette.staticfiles import StaticFiles
from starlette.requests import Request as StarletteRequest
from pathlib import Path
from shiny import App, render, ui, reactive, Inputs, Outputs, Session, req
import webbrowser
import nest_asyncio
import uvicorn
import signal
import os
import sys
import tempfile
import pyrsm as rsm
from pyrsm.utils import ifelse
import pyrsm.radiant.utils as ru
import pyrsm.radiant.model.model_utils as mu


def ui_summary(self):
    return ui.panel_conditional(
        "input.tabs == 'Summary'",
        ui.panel_well(
            ui.input_radio_buttons(
                id="var_type",
                label="Variable type:",
                selected=self.state.get("var_type", "categorical"),
                choices=["categorical", "numeric"],
                inline=True,
            ),
            ui.panel_conditional(
                "input.var_type == 'categorical'",
                ui.output_ui("ui_cvar1"),
                ui.output_ui("ui_cvar2"),
            ),
            ui.panel_conditional(
                "input.var_type == 'numeric'",
                ui.output_ui("ui_nvar1"),
                ui.output_ui("ui_nvar2"),
            ),
            ui.output_ui("ui_combos"),
            ui.input_select(
                id="alt_hyp",
                label="Alternative hypothesis:",
                selected=self.state.get("alt_hyp", "two-sided"),
                choices={
                    "two-sided": "Two sided",
                    "greater": "Greater than",
                    "less": "Less than",
                },
            ),
            ui.input_checkbox(
                id="extra",
                label="Show extra statistics:",
                value=self.state.get("extra", False),
            ),
            ui.panel_conditional(
                "input.extra == true",
                ui.input_slider(
                    id="conf",
                    label="Confidence level:",
                    min=0,
                    max=1,
                    value=self.state.get("conf", 0.95),
                ),
            ),
            ui.input_radio_buttons(
                id="sample_type",
                label="Sample Type:",
                selected=self.state.get("sample_type", "independent"),
                choices={
                    "independent": "independent",
                    "paired": "paired",
                },
                inline=True,
            ),
            ui.input_radio_buttons(
                id="adjust",
                label="Multiple comp. adjustment:",
                selected=self.state.get("adjust", None),
                choices={
                    "None": "None",
                    "bonferroni": "Bonferroni",
                },
                inline=True,
            ),
            ui.input_radio_buttons(
                id="test_type",
                label="Test type:",
                selected=self.state.get("test_type", "t-test"),
                choices={
                    "t-test": "t-test",
                    "wilcox": "Wilcox",
                },
                inline=True,
            ),
        ),
    )


choices = {
    "None": "None",
    "scatter": "Scatter plot",
    "density": "Density plot",
    "box": "Box plot",
    "bar": "Bar chart",
}


def plots_extra(self):
    return (
        ui.panel_conditional(
            "input.plots == 'scatter'",
            ui.input_select(
                "nobs",
                "Number of data points plotted:",
                selected=self.state.get("nobs", "1,000"),
                choices={1_000: "1,000", -1: "All"},
            ),
        ),
    )


class basics_compare_means:
    def __init__(
        self, datasets: dict, descriptions=None, state=None, code=True, navbar=None
    ) -> None:
        ru.init(
            self,
            datasets,
            descriptions=descriptions,
            state=state,
            code=code,
            navbar=navbar,
        )

    def shiny_ui(self, request: StarletteRequest):
        if self.state.get("__pending_changes__", False):
            # from https://gist.github.com/jcheng5/88cb05e39c44704ae89e6559130e7f80
            return ui.tags.html(
                ui.tags.meta({"http-equiv": "refresh", "content": "0"}),
                ui.tags.script("""window.Shiny.createSocket = function() {};"""),
            )

        return ui.page_navbar(
            ru.head_content(),
            ui.nav_panel(
                "Basics > Compare means",
                ui.row(
                    ui.column(
                        3,
                        ru.ui_data(self),
                        ui_summary(self),
                        ru.ui_plot(self, choices, plots_extra(self)),
                    ),
                    ui.column(8, ru.ui_main_basics(self)),
                ),
            ),
            self.navbar,
            ru.ui_help(
                "https://github.com/vnijs/pyrsm/blob/main/examples/basics/basics-compare-means.ipynb",
                "Compare means example notebook",
            ),
            ru.ui_stop(),
            title="Radiant for Python",
            inverse=False,
            id="navbar_id",
        )

    def shiny_server(self, input: Inputs, output: Outputs, session: Session):
        # --- section standard for all apps ---
        get_data = ru.make_data_elements(self, input, output, session)

        def update_state():
            with reactive.isolate():
                ru.dct_update(self, input)

        session.on_ended(update_state)

        # issue when launching external browser from browser inside VS Code
        # def on_flushed():
        #     self.state["__pending_changes__"] = True

        # session.on_flushed(on_flushed, once=False)

        # --- section unique to each app ---
        @output(id="ui_cvar1")
        @render.ui
        def ui_cvar1():
            isCat = get_data()["var_types"]["isCat"]
            return ui.input_select(
                id="cvar1",
                label="Select a categorical variable:",
                selected=self.state.get("cvar1", None),
                choices=isCat,
            )

        @output(id="ui_cvar2")
        @render.ui
        def ui_cvar2():
            isNum = get_data()["var_types"]["isNum"].copy()
            return ui.input_select(
                id="cvar2",
                label="Numeric Variable:",
                selected=self.state.get("cvar2", None),
                choices=isNum,
            )

        @output(id="ui_nvar1")
        @render.ui
        def ui_nvar1():
            isNum = get_data()["var_types"]["isNum"]
            return ui.input_select(
                id="nvar1",
                label="Select a numeric variable:",
                selected=self.state.get("nvar1", None),
                choices=isNum,
            )

        @output(id="ui_nvar2")
        @render.ui
        def ui_nvar2():
            isNum = get_data()["var_types"]["isNum"].copy()
            if input.nvar1() is not None:
                del isNum[input.nvar1()]
            return ui.input_select(
                id="nvar2",
                label="Numeric Variable:",
                selected=self.state.get("nvar2", None),
                choices=isNum,
                multiple=True,
                size=min(8, len(isNum)),
            )

        def combo_choices():
            data = get_data()["data"].copy()
            if input.var_type() == "categorical":
                levels = data[input.cvar1()].astype("category").cat.categories
            else:
                data = data.loc[:, [input.nvar1()] + list(input.nvar2())].melt()
                levels = data["variable"].unique()

            return list(ru.iterms(levels))

        @output(id="ui_combos")
        @render.ui
        def ui_combos():
            req(input.var_type())
            choices = combo_choices()
            return ui.input_select(
                id="comb",
                label="Choose combinations:",
                selected=self.state.get("comb", None),
                choices=choices,
                multiple=True,
                size=min(3, len(choices)),
            )

        def estimation_code():
            data_name, code = (get_data()[k] for k in ["data_name", "code"])
            if input.var_type() == "categorical":
                var1 = input.cvar1()
                var2 = input.cvar2()
            else:
                var1 = input.nvar1()
                var2 = input.nvar2()

            args = {
                "data": f"""{{"{data_name}": {data_name}}}""",
                "var1": var1,
                "var2": var2,
                "comb": list(input.comb()),
                "alt_hyp": input.alt_hyp(),
                "conf": input.conf(),
                "sample_type": input.sample_type(),
                "adjust": ifelse(input.adjust() == "None", None, input.adjust()),
                "test_type": input.test_type(),
            }

            args_string = ru.drop_default_args(args, rsm.basics.compare_means)
            return f"""rsm.basics.compare_means({args_string})""", code

        show_code, estimate = mu.make_estimate(
            self,
            input,
            output,
            get_data,
            fun="basics.compare_means",
            ret="cm",
            ec=estimation_code,
            run=False,
            debug=True,
        )

        def summary_code():
            args = {"extra": input.extra()}
            args_string = ru.drop_default_args(args, rsm.basics.compare_means.summary)
            return f"""cm.summary({args_string})"""

        mu.make_summary(
            self,
            input,
            output,
            session,
            show_code,
            estimate,
            ret="cm",
            sum_fun=rsm.basics.compare_means.summary,
            sc=summary_code,
        )

        def plot_code():
            args = {"plots": input.plots(), "nobs": input.nobs()}
            if input.plots() != "scatter":
                args["nobs"] = None
            args_string = ru.drop_default_args(
                args, rsm.basics.compare_means.plot, ignore=["nobs"]
            )
            return f"""cm.plot({args_string})"""

        mu.make_plot(
            self,
            input,
            output,
            session,
            show_code,
            estimate,
            ret="cm",
            pc=plot_code,
        )

        # --- section standard for all apps ---
        # stops returning code if moved to utils
        @reactive.Effect
        @reactive.event(input.stop, ignore_none=True)
        async def stop_app():
            rsm.md(f"```python\n{self.stop_code}\n```")
            await session.app.stop()
            os.kill(os.getpid(), signal.SIGTERM)


def compare_means(
    data_dct: dict = None,
    descriptions_dct: dict = None,
    state: dict = None,
    code: bool = True,
    host: str = "",
    port: int = 8000,
    log_level: str = "warning",
    debug: bool = False,
):
    """
    Launch a Radiant-for-Python app for compare means hypothesis testing
    """
    host = ru.set_host(host)
    if data_dct is None:
        data_dct, descriptions_dct = ru.get_dfs(pkg="basics", name="salary")
    rc = basics_compare_means(data_dct, descriptions_dct, state=state, code=code)
    nest_asyncio.apply()
    webbrowser.open(f"http://{host}:{port}")
    print(f"Listening on http://{host}:{port}")
    ru.message()

    # redirect stdout and stderr to the temporary file
    if not debug:
        org_stdout = sys.stdout
        org_stderr = sys.stderr
        temp = tempfile.NamedTemporaryFile()
        temp_file = open(temp.name, "w")
        sys.stdout = temp_file
        sys.stderr = temp_file

    app = App(rc.shiny_ui, rc.shiny_server)
    www_dir = Path(__file__).parent.parent.parent / "radiant" / "www"
    app_static = StaticFiles(directory=www_dir, html=False)

    routes = [
        Mount("/www", app=app_static),
        Mount("/", app=app),
    ]

    uvicorn.run(
        Starlette(debug=debug, routes=routes),
        host=host,
        port=port,
        log_level=log_level,
    )

    if not debug:
        sys.stdout = org_stdout
        sys.stderr = org_stderr
        temp_file.close()


if __name__ == "__main__":
    compare_means(debug=False)



import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from typing import Optional


class central_limit_theorem:
    def __init__(
        self,
        dist: str,
        sample_size: int,
        num_samples: int,
        num_bins: int,
        figsize: Optional[tuple[float, float]] = (10, 10),
        **params: float,
    ) -> None:
        self.dist = dist
        self.sample_size = sample_size
        self.num_samples = num_samples
        self.num_bins = max(1, min(num_bins, 50))
        self.figsize = figsize
        self.params = params

    def simulate(self) -> None:
        if self.dist == "normal":
            self.simulate_normal()
        elif self.dist == "binomial":
            self.simulate_binomial()
        elif self.dist == "uniform":
            self.simulate_uniform()
        elif self.dist == "exponential":
            self.simulate_exponential()
        else:
            print("Invalid distribution")

    def simulate_normal(self) -> None:
        mean = self.params["mean"]
        sd = self.params["sd"]
        samples = [
            np.random.normal(loc=mean, scale=sd, size=self.sample_size)
            for _ in range(self.num_samples)
        ]

        self.plot_distribution(samples)

    def simulate_binomial(self) -> None:
        size = self.params["size"]
        prob = self.params["prob"]

        samples = [
            np.random.binomial(n=size, p=prob, size=self.sample_size)
            for _ in range(self.num_samples)
        ]

        self.plot_distribution(samples)

    def simulate_uniform(self) -> None:
        minimum = self.params["min"]
        maximum = self.params["max"]

        samples = [
            np.random.uniform(low=minimum, high=maximum, size=self.sample_size)
            for _ in range(self.num_samples)
        ]

        self.plot_distribution(samples)

    def simulate_exponential(self) -> None:
        rate = self.params["rate"]

        samples = [
            np.random.exponential(scale=rate, size=self.sample_size)
            for _ in range(self.num_samples)
        ]

        self.plot_distribution(samples)

    def plot_distribution(self, samples: list[np.ndarray]) -> None:
        sample_means = [np.mean(sample) for sample in samples]

        _, axes = plt.subplots(2, 2, figsize=self.figsize)
        plt.subplots_adjust(wspace=0.2, hspace=0.2)

        plt.axes(axes[0][0])
        self._plot_distribution(x=samples[0], x_label="Histogram of sample #1").plot()

        plt.axes(axes[0][1])
        self._plot_distribution(
            x=samples[-1], x_label="Histogram of sample #" + str(self.num_samples)
        ).plot()

        plt.axes(axes[1][0])
        self._plot_distribution(
            x=sample_means, x_label="Histogram of sample means"
        ).plot()

        plt.axes(axes[1][1])
        axes[1][1].set_ylabel("y")
        self._plot_distribution(
            x=sample_means, x_label="Density of sample means", density_plot=True
        ).plot()

    def _plot_distribution(
        self, x: list[np.ndarray], x_label: str, density_plot: bool = False
    ) -> matplotlib.axes.Axes:
        stat = "count"
        data = {x_label: x}

        data = pd.DataFrame(data)
        if density_plot:
            return sns.kdeplot(data=data, x=x_label, fill=True)

        return sns.histplot(
            data=data,
            x=x_label,
            stat=stat,
            bins=self.num_bins,
        )


if __name__ == "__main__":
    clt = central_limit_theorem(
        dist="normal",
        sample_size=1000,
        num_samples=1000,
        num_bins=30,
        figsize=(10, 10),
        mean=0,
        sd=1,
    )
    clt.simulate()

    clt = central_limit_theorem(
        dist="binomial",
        sample_size=1000,
        num_samples=1000,
        num_bins=30,
        figsize=(10, 10),
        size=10,
        prob=0.1,
    )

    clt.simulate()



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import polars as pl
from scipy import stats
from pyrsm.utils import ifelse, check_dataframe
from typing import Union


class cross_tabs:
    """
    Calculate a Chi-square test between two categorical variables contained
    in a Pandas dataframe

    Parameters
    ----------
    data : Pandas or Polars dataframe with categorical variables or a
        dictionary with a single dataframe as the value and the
        name of the dataframe as the key
    var1: String; Name of the first categorical variable
    var2: String; Name of the second categorical variable

    Returns
    -------
    Cross object with several attributes
    data: Original dataframe
    var1: Name of the first categorical variable
    var2: Name of the second categorical variable
    observed: Dataframe of observed frequencies
    expected: Dataframe of expected frequencies
    expected_low: List with number of cells with expected values < 5
        and the total number of cells
    chisq: Dataframe of chi-square values for each cell
    dev_std: Dataframe of standardized deviations from the expected table
    perc_row: Dataframe of observation percentages conditioned by row
    perc_col: Dataframe of observation percentages conditioned by column
    perc: Dataframe of observation percentages by the total number of observations

    Examples
    --------
    import pyrsm as rsm
    newspaper, newspapar_description = rsm.load_data(pkg="basics", name="newspaper")
    ct = rsm.cross_tabs(newspaper, "Income", "Newspaper")
    ct.expected
    """

    def __init__(
        self,
        data: pd.DataFrame | pl.DataFrame | dict[str, pd.DataFrame | pl.DataFrame],
        var1: str,
        var2: str,
    ) -> None:
        if isinstance(data, dict):
            self.name = list(data.keys())[0]
            self.data = data[self.name]
        else:
            self.data = data
            self.name = "Not provided"

        self.data = check_dataframe(self.data)
        self.var1 = var1
        self.var2 = var2

        self.observed = pd.crosstab(
            self.data[var1], columns=self.data[var2], margins=True, margins_name="Total"
        )
        self.chisq_test = stats.chi2_contingency(
            self.observed.drop(columns="Total").drop("Total", axis=0), correction=False
        )
        expected = pd.DataFrame(self.chisq_test[3])
        self.expected_low = [
            (expected < 5).sum().sum(),
            expected.shape[0] * expected.shape[0],
        ]
        expected["Total"] = expected.sum(axis=1)
        expected = pd.concat(
            [expected, expected.sum().to_frame().T], ignore_index=True, axis=0
        ).set_index(self.observed.index)
        expected.columns = self.observed.columns
        self.expected = expected
        chisq = (self.observed - self.expected) ** 2 / self.expected
        chisq["Total"] = chisq.sum(axis=1)
        chisq.loc["Total", :] = chisq.sum()
        self.chisq = chisq
        self.dev_std = (
            ((self.observed - self.expected) / np.sqrt(self.expected))
            .drop(columns="Total")
            .drop("Total", axis=0)
        )
        self.perc_row = self.observed.div(self.observed["Total"], axis=0)
        self.perc_col = self.observed.div(self.observed.loc["Total", :], axis=1)
        self.perc = self.observed / self.observed.loc["Total", "Total"]

    def summary(
        self, output: list[str] = ["observed", "expected"], dec: int = 2
    ) -> None:
        """
        Print different output tables for a cross_tabs object

        Parameters
        ----------
        output : list of tables to show
            Options include "observed" (observed frequencies),
            "expected" (expected frequencies), "chisq" (chi-square values)
            for each cell, "dev_std" (standardized deviations from expected)
            "perc_row" (percentages conditioned by row), "perc_col"
            (percentages conditioned by column), "perc" (percentages by the
            total number of observations). The default value is ["observed", "expected"]
        dec : int
            Number of decimal places to use in rounding

        Examples
        --------
        import pyrsm as rsm
        newspaper, newspapar_description = rsm.load_data(pkg="basics", name="newspaper")
        ct = rsm.cross_tabs(newspaper)
        ct.summary()
        """

        pd.set_option("display.max_columns", 20)
        pd.set_option("display.max_rows", 20)

        output = ifelse(isinstance(output, str), [output], output)
        prn = f"""
Cross-tabs
Data     : {self.name}
Variables: {self.var1}, {self.var2}
Null hyp : There is no association between {self.var1} and {self.var2}
Alt. hyp : There is an association between {self.var1} and {self.var2}
"""
        if "observed" in output:
            prn += f"""
Observed:

{self.observed.map(lambda x: "{:,}".format(x))}
"""
        if "expected" in output:
            prn += f"""
Expected: (row total x column total) / total

{self.expected.round(dec).map(lambda x: "{:,}".format(x))}
"""
        if "chisq" in output:
            prn += f"""
Contribution to chi-squared: (o - e)^2 / e

{self.chisq.round(dec).map(lambda x: "{:,}".format(x))}
"""

        if "dev_std" in output:
            prn += f"""
Deviation standardized: (o - e) / sqrt(e)

{self.dev_std.round(dec).map(lambda x: "{:,}".format(x))}
"""

        if "perc_row" in output:
            prn += f"""
Row percentages:

{self.perc_row.transform(lambda x: (100*x).round(dec).astype(str) + "%")}
"""
        if "perc_col" in output:
            prn += f"""
Column percentages:

{self.perc_col.transform(lambda x: (100*x).round(dec).astype(str) + "%")}
"""
        if "perc" in output:
            prn += f"""
Percentages:

{self.perc.transform(lambda x: (100*x).round(dec).astype(str) + "%")}
"""
        prn += f"""
Chi-squared: {round(self.chisq_test[0], dec)} df({int(self.chisq_test[2])}), p.value {ifelse(self.chisq_test[1] < 0.001, "< .001", round(self.chisq_test[1], dec))}
{100 * round(self.expected_low[0] / self.expected_low[1], dec)}% of cells have expected values below 5
"""
        print(prn)

        pd.reset_option("display.max_columns")
        pd.reset_option("display.max_rows")

    def plot(self, plots: list[str] = "perc_col", **kwargs) -> None:
        """
        Plot of correlations between numeric variables in a Pandas dataframe

        Parameters
        ----------
        plots : list of tables to show
            Options include "observed" (observed frequencies),
            "expected" (expected frequencies), "chisq" (chi-square values)
            for each cell, "dev_std" (standardized deviations from expected)
            "perc_row" (percentages conditioned by row), "perc_col"
            (percentages conditioned by column), "perc" (percentages by the
            total number of observations). The default value is ["observed", "expected"]
        **kwargs : Named arguments to be passed to pandas plotting functions

        Examples
        --------
        import pyrsm as rsm
        newspaper, newspapar_description = rsm.load_data(pkg="basics", name="newspaper")
        ct = rsm.cross_tabs(newspaper, "Income", "Newspaper")
        ct.plot()
        """
        plots = ifelse(isinstance(plots, str), [plots], plots)

        args = {"rot": False}
        if "observed" in plots:
            tab = (
                self.observed.transpose()
                .drop(columns="Total")
                .drop("Total", axis=0)
                .apply(lambda x: x * 100 / sum(x), axis=1)
            )
            args["title"] = "Observed frequencies"
            args.update(**kwargs)
            fig = tab.plot.bar(stacked=True, **args)
        if "expected" in plots:
            tab = (
                self.expected.transpose()
                .drop(columns="Total")
                .drop("Total", axis=0)
                .apply(lambda x: x * 100 / sum(x), axis=1)
            )
            args["title"] = "Expected frequencies"
            args.update(**kwargs)
            fig = tab.plot.bar(stacked=True, **args)
        if "chisq" in plots:
            tab = self.chisq.transpose().drop(columns="Total").drop("Total", axis=0)
            args["title"] = "Contribution to chi-squared statistic"
            args.update(**kwargs)
            fig = tab.plot.bar(**args)
        if "dev_std" in plots:
            tab = self.dev_std.transpose()
            args["title"] = "Deviation standardized"
            args.update(**kwargs)
            fig, ax = plt.subplots()
            tab.plot.bar(**args, ax=ax)
            ax.axhline(y=1.96, color="black", linestyle="--")
            ax.axhline(y=1.64, color="black", linestyle="--")
            ax.axhline(y=0, color="black")
            ax.axhline(y=-1.96, color="black", linestyle="--")
            ax.axhline(y=-1.64, color="black", linestyle="--")
            ax.annotate("95%", xy=(0, 2.1), va="bottom", ha="center")
            ax.annotate("90%", xy=(0, 1.4), va="top", ha="center")
        if "perc_col" in plots:
            tab = self.perc_col.transpose().drop(columns="Total").drop("Total", axis=0)
            args["title"] = "Column percentages"
            args.update(**kwargs)
            fig = tab.plot.bar(**args)
        if "perc_row" in plots:
            tab = self.perc_row.transpose().drop(columns="Total").drop("Total", axis=0)
            args["title"] = "Row percentages"
            args.update(**kwargs)
            fig = tab.plot.bar(**args)
        if "perc" in plots:
            tab = self.perc.transpose().drop(columns="Total").drop("Total", axis=0)
            args["title"] = "Table percentages"
            args.update(**kwargs)
            fig = tab.plot.bar(**args)



from starlette.applications import Starlette
from starlette.routing import Mount
from starlette.staticfiles import StaticFiles
from starlette.requests import Request as StarletteRequest
from pathlib import Path
from shiny import App, render, ui, reactive, Inputs, Outputs, Session, req
import webbrowser
import nest_asyncio
import uvicorn
import signal
import os
import sys
import tempfile
import pyrsm as rsm
from pyrsm.utils import ifelse
import pyrsm.radiant.utils as ru
import pyrsm.radiant.model.model_utils as mu


def ui_summary(self):
    return ui.panel_conditional(
        "input.tabs == 'Summary'",
        ui.panel_well(
            ui.output_ui("ui_var1"),
            ui.output_ui("ui_var2"),
            ui.output_ui("ui_lev"),
            ui.output_ui("ui_combos"),
            ui.input_select(
                id="alt_hyp",
                label="Alternative hypothesis:",
                selected=self.state.get("alt_hyp", "two-sided"),
                choices={
                    "two-sided": "Two sided",
                    "greater": "Greater than",
                    "less": "Less than",
                },
            ),
            ui.input_checkbox(
                id="extra",
                label="Show extra statistics:",
                value=self.state.get("extra", False),
            ),
            ui.panel_conditional(
                "input.extra == true",
                ui.input_slider(
                    id="conf",
                    label="Confidence level:",
                    min=0,
                    max=1,
                    value=self.state.get("conf", 0.95),
                ),
            ),
            ui.input_radio_buttons(
                id="adjust",
                label="Multiple comp. adjustment:",
                selected=self.state.get("adjust", None),
                choices={
                    "None": "None",
                    "bonferroni": "Bonferroni",
                },
                inline=True,
            ),
        ),
    )


choices = {
    "None": "None",
    "bar": "Bar chart",
    "dodge": "Dodge plot",
}


class basics_compare_props:
    def __init__(
        self, datasets: dict, descriptions=None, state=None, code=True, navbar=None
    ) -> None:
        ru.init(
            self,
            datasets,
            descriptions=descriptions,
            state=state,
            code=code,
            navbar=navbar,
        )

    def shiny_ui(self, request: StarletteRequest):
        return ui.page_navbar(
            ru.head_content(),
            ui.nav_panel(
                "Basics > Compare props",
                ui.row(
                    ui.column(
                        3,
                        ru.ui_data(self),
                        ui_summary(self),
                        ru.ui_plot(self, choices, None),
                    ),
                    ui.column(8, ru.ui_main_basics(self)),
                ),
            ),
            self.navbar,
            ru.ui_help(
                "https://github.com/vnijs/pyrsm/blob/main/examples/basics/basics-compare-props.ipynb",
                "Compare props example notebook",
            ),
            ru.ui_stop(),
            title="Radiant for Python",
            inverse=False,
            id="navbar_id",
        )

    def shiny_server(self, input: Inputs, output: Outputs, session: Session):
        # --- section standard for all apps ---
        get_data = ru.make_data_elements(self, input, output, session)

        def update_state():
            with reactive.isolate():
                ru.dct_update(self, input)

        session.on_ended(update_state)

        # --- section unique to each app ---
        @output(id="ui_var1")
        @render.ui
        def ui_var1():
            isCat = get_data()["var_types"]["isCat"]
            return ui.input_select(
                id="var1",
                label="Select a categorical variable:",
                selected=self.state.get("var1", None),
                choices=isCat,
            )

        @output(id="ui_var2")
        @render.ui
        def ui_var2():
            isBin = get_data()["var_types"]["isBin"].copy()
            return ui.input_select(
                id="var2",
                label="Categorical Variable (binary):",
                selected=self.state.get("var2", None),
                choices=isBin,
            )

        @output(id="ui_lev")
        @render.ui
        def ui_lev():
            levs = list(get_data()["data"][input.var2()].unique())
            return ui.input_select(
                id="lev",
                label="Choose level:",
                selected=self.state.get("lev", levs[0]),
                choices=levs,
            )

        def combo_choices():
            data = get_data()["data"].copy()
            levels = data[input.var1()].astype("category").cat.categories
            return list(ru.iterms(levels))

        @output(id="ui_combos")
        @render.ui
        def ui_combos():
            choices = combo_choices()
            return ui.input_select(
                id="comb",
                label="Choose combinations:",
                selected=self.state.get("comb", None),
                choices=choices,
                multiple=True,
                size=min(3, len(choices)),
            )

        def estimation_code():
            data_name, code = (get_data()[k] for k in ["data_name", "code"])
            args = {
                "data": f"""{{"{data_name}": {data_name}}}""",
                "var1": input.var1(),
                "var2": input.var2(),
                "lev": input.lev(),
                "comb": list(input.comb()),
                "alt_hyp": input.alt_hyp(),
                "conf": input.conf(),
                "adjust": ifelse(input.adjust() == "None", None, input.adjust()),
            }

            args_string = ru.drop_default_args(args, rsm.basics.compare_props)
            return f"""rsm.basics.compare_props({args_string})""", code

        show_code, estimate = mu.make_estimate(
            self,
            input,
            output,
            get_data,
            fun="basics.compare_props",
            ret="cp",
            ec=estimation_code,
            run=False,
            debug=True,
        )

        def summary_code():
            args = {"extra": input.extra()}
            args_string = ru.drop_default_args(args, rsm.basics.compare_props.summary)
            return f"""cp.summary({args_string})"""

        mu.make_summary(
            self,
            input,
            output,
            session,
            show_code,
            estimate,
            ret="cp",
            sum_fun=rsm.basics.compare_props.summary,
            sc=summary_code,
        )

        def plot_code():
            args = {"plots": input.plots()}
            args_string = ru.drop_default_args(args, rsm.basics.compare_props.plot)
            return f"""cp.plot({args_string})"""

        mu.make_plot(
            self,
            input,
            output,
            session,
            show_code,
            estimate,
            ret="cp",
            pc=plot_code,
        )

        # --- section standard for all apps ---
        # stops returning code if moved to utils
        @reactive.Effect
        @reactive.event(input.stop, ignore_none=True)
        async def stop_app():
            rsm.md(f"```python\n{self.stop_code}\n```")
            await session.app.stop()
            os.kill(os.getpid(), signal.SIGTERM)


def compare_props(
    data_dct: dict = None,
    descriptions_dct: dict = None,
    state: dict = None,
    code: bool = True,
    host: str = "",
    port: int = 8000,
    log_level: str = "warning",
    debug: bool = False,
):
    """
    Launch a Radiant-for-Python app for compare props hypothesis testing
    """
    host = ru.set_host(host)
    if data_dct is None:
        data_dct, descriptions_dct = ru.get_dfs(pkg="data", name="titanic")
    rc = basics_compare_props(data_dct, descriptions_dct, state=state, code=code)
    nest_asyncio.apply()
    webbrowser.open(f"http://{host}:{port}")
    print(f"Listening on http://{host}:{port}")
    ru.message()

    # redirect stdout and stderr to the temporary file
    if not debug:
        org_stdout = sys.stdout
        org_stderr = sys.stderr
        temp = tempfile.NamedTemporaryFile()
        temp_file = open(temp.name, "w")
        sys.stdout = temp_file
        sys.stderr = temp_file

    app = App(rc.shiny_ui, rc.shiny_server)
    www_dir = Path(__file__).parent.parent.parent / "radiant" / "www"
    app_static = StaticFiles(directory=www_dir, html=False)

    routes = [
        Mount("/www", app=app_static),
        Mount("/", app=app),
    ]

    uvicorn.run(
        Starlette(debug=debug, routes=routes),
        host=host,
        port=port,
        log_level=log_level,
    )

    if not debug:
        sys.stdout = org_stdout
        sys.stderr = org_stderr
        temp_file.close()


if __name__ == "__main__":
    compare_props(debug=False)



import numpy as np
import pandas as pd
import sys


def bincode(x, breaks, right=True, include_lowest=True):
    """
    Split a numeric series into a number of bins and return a series of bin numbers

    Port of R's .bincode function

    Parameters
    ----------
    x : List, pandas series, or numpy array
        Numeric variable to bin
    breaks : int
        Numpy array of two or more cut points, sorted in increasing order
    right: bool
        Should intervals be closed on the right (and open on the left) or vice versa
    include_lowest : bool
        Should an ‘x[i]’ equal to the lowest (or highest, for right = FALSE) ‘breaks’ value be included in the first (or last) bin

    Returns
    -------
    Numpy array with bin numbers for each numeric value in x

    Examples
    --------
    bincode(np.arange(10), np.array([3, 5, 8]))
    """
    n = len(x)
    if any(np.isnan(x)):
        code = np.array([0.0] * n)
    else:
        code = np.array([0] * n)
    nb = len(breaks)
    nb1 = nb - 1
    lft = right is False

    # check if breaks are sorted
    try:
        if sum(breaks[1:] > breaks[-1]) > 0:
            raise ValueError
    except ValueError:
        print("Error: Breaks are not sorted")
        sys.exit(1)

    for i in range(n):
        if np.isnan(x[i]):
            code[i] = np.NaN
            print(f"Entry {i} is a missing value")
        else:
            lo = 0
            hi = nb1
            if (
                x[i] < breaks[lo]
                or breaks[hi] < x[i]
                or (x[i] == breaks[np.where(lft, hi, lo)] and include_lowest is False)
            ):
                next
            else:
                while hi - lo >= 2:
                    new = int(round((hi + lo) / 2, 0))
                    if x[i] > breaks[new] or (lft and x[i] == breaks[new]):
                        lo = new
                    else:
                        hi = new
                code[i] = lo + 1

    return code


def xtile(x, n=5, rev=False):
    """
    Split a numeric series into a number of bins and return a series of bin numbers

    Parameters
    ----------
    x : List, pandas series, or numpy array
        Numeric variable to bin
    n :	int
        Number of bins to create
    rev	: bool
        Reverse the order of the bin numbers (False or True)

    Returns
    -------
    Numpy array with bin numbers for each numeric value in x

    Examples
    --------
    xtile(np.arange(10), 5)
    xtile(np.arange(10), 5, rev=True)
    """

    x = np.array(x)
    breaks = np.quantile(x[~np.isnan(x)], np.arange(0, n + 1) / n)
    breaks
    if len(np.unique(breaks)) == len(breaks):
        bins = pd.cut(x, breaks, include_lowest=True, labels=False) + 1
    else:
        bins = bincode(x, breaks)

    if rev is True:
        bins = (n + 1) - bins

    return bins



import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
import pandas as pd
import polars as pl
from scipy import stats
from pyrsm.model import sig_stars
from pyrsm.utils import ifelse, check_dataframe
from typing import Literal
from statsmodels.stats import multitest
import pyrsm.basics.utils as bu
import pyrsm.radiant.utils as ru


class compare_means:
    """
    A class to perform comparison of means hypothesis testing. See the notebook 
    linked below for a worked example, including the web UI:

    https://github.com/vnijs/pyrsm/blob/main/examples/basics-compare-means.ipynb

    Attributes
    ----------
    data : pd.DataFrame | pl.DataFrame
        The input data for the hypothesis test as a Pandas or Polars DataFrame.
    var1 : str
        The first variable/column name to test.
    var2 : str
        The second variable/column name to test.
    alt_hyp : str
        The alternative hypothesis ('two-sided', 'greater', 'less').
    conf : float
        The confidence level for the test.
    comp_value : float
        The comparison value for the test.
    t_val : float
        The t-statistic value.
    p_val : float
        The p-value of the test.
    ci : tuple
        The confidence interval of the test.
    mean1 : float
        The mean of the first variable.
    mean2 : float
        The mean of the second variable.
    n1 : int
        The number of observations for the first variable.
    n2 : int
        The number of observations for the second variable.
    sd1 : float
        The standard deviation of the first variable.
    sd2 : float
        The standard deviation of the second variable.
    se : float
        The standard error of the difference between the means.
    me : float
        The margin of error.
    diff : float
        The difference between the means of the two variables.
    df : int
        The degrees of freedom.

    Methods
    -------
    __init__(data, var1, var2, alt_hyp='two-sided', conf=0.95, comp_value=0)
        Initializes the compare_means class with the provided data and parameters.
    summary(dec=3)
        Prints a summary of the hypothesis test.
    plot()
        Plots the results of the hypothesis test.

    Examples
    --------

    import pandas as pd
    import pyrsm as rsm
    salary, salary_description = rsm.load_data(pkg="basics", name="salary")
    cm = rsm.basics.compare_means({"salary": salary}, var1="rank", var2="salary", alt_hyp="less")
    cm.summary()
    """

    def __init__(
        self,
        data: pd.DataFrame | pl.DataFrame | dict[str, pd.DataFrame | pl.DataFrame],
        var1: str,
        var2: str,
        comb: list[tuple[str, str]] = [],
        alt_hyp: Literal["two-sided", "greater", "less"] = "two-sided",
        conf: float = 0.95,
        sample_type: Literal["independent", "paired"] = "independent",
        adjust: Literal[None, "bonferroni"] = None,
        test_type: Literal["t-test", "wilcox"] = "t-test",
    ):
        """
        Constructs all the necessary attributes for the compare_means object.

        Parameters
        ----------
        data : pd.DataFrame | pl.DataFrame
            The input data for the hypothesis test as a Pandas or Polars DataFrame. If a dictionary is provided, the key should be the name of the dataframe.
        var1 : str
            The first variable/column name to include in the test. This variable can be numeric or categorical. If it is categorical, the hypothesis test will be performed for each level of the variable.
        var2 : str
            The second variable/column name or names to include in the test. These variables must be numeric. If multiple variables are provided, the hypothesis test will be performed for each combination of variables. If var1 is categorical, only one variable can be provided.
        alt_hyp : str, optional
            The alternative hypothesis ('two-sided', 'greater', 'less') (default is 'two-sided').
        conf : float, optional
            The confidence level for the test (default is 0.95).
        sample_type : str
            The type of samples ('independent' or 'paired') (default is 'independent').An example of paired samples is when the same subjects are measured at two different time points.
        adjust : str, optional
            Adjustment for multiple testing (None or 'bonferroni' or other options provided but statsmodels.stats.multitest.multipletests).
        test_type : str
            The type of test ('t-test' or 'wilcox') (default is 't-test'). The key difference between the two tests is that the t-test assumes that the data is normally distributed, while the Wilcoxon test does not.
        """
        if isinstance(data, dict):
            self.name = list(data.keys())[0]
            self.data = data[self.name]
        else:
            self.data = data
            self.name = "Not provided"

        self.data = check_dataframe(self.data)
        self.var1 = var1
        self.var2 = var2
        self.comb = comb

        var1_series = self.data[self.var1]
        if pd.api.types.is_numeric_dtype(var1_series):
            var2 = ifelse(
                isinstance(var2, str),
                [var2],
                ifelse(isinstance(var2, tuple), list(var2), var2),
            )
            if (
                len(self.var2) > 1
                or len([v for v in self.comb if self.var1 in v]) > 0
                or var1_series.nunique() > 10
            ):
                self.data = self.data.loc[:, [self.var1] + var2].melt()
                self.var1 = "variable"
                self.var2 = "value"

        self.data[self.var1] = self.data[self.var1].astype("category")
        self.levels = list(self.data[self.var1].cat.categories)

        if len(self.comb) == 0:
            self.comb = ru.iterms(self.levels)

        self.alt_hyp = alt_hyp
        self.conf = conf
        self.alpha = 1 - self.conf
        self.sample_type = sample_type
        self.adjust = adjust
        self.test_type = test_type

        def welch_dof(v1: str, v2: str) -> float:
            # stats.ttest_ind uses Welch's t-test when equal_var=False
            # but does not return the degrees of freedom
            x = self.data.loc[self.data[self.var1] == v1, self.var2]
            y = self.data.loc[self.data[self.var1] == v2, self.var2]
            if x.size == 0 or y.size == 0:  # address division by zero
                return np.nan
            dof = (x.var() / x.size + y.var() / y.size) ** 2 / (
                (x.var() / x.size) ** 2 / (x.size - 1)
                + (y.var() / y.size) ** 2 / (y.size - 1)
            )

            return dof

        descriptive_stats = []
        for lev in self.levels:
            subset = self.data.loc[self.data[self.var1] == lev, self.var2]
            mean = np.nanmean(subset)
            n_missing = subset.isna().sum()
            n = len(subset) - n_missing
            sd = subset.std()
            se = subset.sem()
            tscore = stats.t.ppf((1 + self.conf) / 2, n - 1)
            me = (tscore * se).real
            row = [lev, mean, n, n_missing, sd, se, me]
            descriptive_stats.append(row)

        self.descriptive_stats = pd.DataFrame(
            descriptive_stats,
            columns=[self.var1, "mean", "n", "n_missing", "sd", "se", "me"],
        )

        if self.alt_hyp == "less":
            alt_hyp_sign = "less than"
        elif self.alt_hyp == "two-sided":
            alt_hyp_sign = "not equal to"
        else:
            alt_hyp_sign = "greater than"

        comp_stats = []
        for c in self.comb:
            v1, v2 = c.split(":")
            null_hyp = f"{v1} = {v2}"
            alt_hyp = f"{v1} {alt_hyp_sign} {v2}"
            diff = np.nanmean(
                self.data[self.data[self.var1] == v1][self.var2]
            ) - np.nanmean(self.data[self.data[self.var1] == v2][self.var2])

            x = self.data.loc[self.data[self.var1] == v1, self.var2]
            y = self.data.loc[self.data[self.var1] == v2, self.var2]
            if x.size != y.size and self.sample_type == "paired":
                raise ValueError(
                    """The two samples must have the same size for a paired
                    sample test. Choose independent samples instead."""
                )

            if self.test_type == "t-test":
                if self.sample_type == "independent":
                    result = stats.ttest_ind(
                        x,
                        y,
                        equal_var=False,
                        nan_policy="omit",
                        alternative=self.alt_hyp,
                    )
                else:
                    result = stats.ttest_rel(
                        x, y, nan_policy="omit", alternative=self.alt_hyp
                    )
            elif self.test_type == "wilcox":
                if self.sample_type == "independent":
                    result = stats.ranksums(x, y, alternative=self.alt_hyp)
                else:
                    result = stats.wilcoxon(
                        x, y, correction=True, alternative=self.alt_hyp
                    )

            t_val, p_val = result.statistic, result.pvalue
            se = diff / t_val
            df = welch_dof(v1, v2)

            if self.alt_hyp == "two-sided":
                tscore = stats.t.ppf((1 + self.conf) / 2, df)
            else:
                tscore = stats.t.ppf(self.conf, df)
            me = (tscore * se).real

            if self.alt_hyp == "less":
                ci = [-np.inf, diff + me]
            elif self.alt_hyp == "two-sided":
                ci = [diff - me, diff + me]
            else:
                ci = [diff - me, np.inf]

            comp_stats.append(
                [
                    null_hyp,
                    alt_hyp,
                    diff,
                    p_val,
                    se,
                    t_val,
                    df,
                    ci[0],
                    ci[1],
                    sig_stars([p_val])[0],
                ]
            )

        cl = bu.ci_label(self.alt_hyp, self.conf)
        self.comp_stats = pd.DataFrame(
            comp_stats,
            columns=[
                "Null hyp.",
                "Alt. hyp.",
                "diff",
                "p.value",
                "se",
                "t.value",
                "df",
                cl[0],
                cl[1],
                "",
            ],
        )

        if self.adjust is not None:
            if self.alt_hyp == "two-sided":
                alpha = self.alpha
            else:
                alpha = self.alpha * 2
            self.comp_stats["p.value"] = multitest.multipletests(
                self.comp_stats["p.value"], method=self.adjust, alpha=alpha
            )[1]
            self.comp_stats[""] = sig_stars(self.comp_stats["p.value"])

    def summary(self, extra=False, dec=3) -> None:
        """
        Prints a summary of the hypothesis test.

        Parameters
        ----------
        extra : bool
            Whether to include additional columns in the output (default is False).
        dec : int, optional
            The number of decimal places to display (default is 3).
        """
        print(f"Pairwise mean comparisons ({self.test_type})")
        print(f"Data      : {self.name}")
        print(f"Variables : {self.var1}, {self.var2}")
        print(f"Samples   : {self.sample_type}")
        print(f"Confidence: {self.conf}")
        print(f"Adjustment: {self.adjust}")

        print(self.descriptive_stats.round(dec).to_string(index=False))

        comp_stats = self.comp_stats.copy()
        if not extra:
            comp_stats = comp_stats.iloc[:, [0, 1, 2, 3, -1]]

        comp_stats["p.value"] = ifelse(
            comp_stats["p.value"] < 0.001, "< .001", round(comp_stats["p.value"], dec)
        )
        print(comp_stats.round(dec).to_string(index=False))
        print("\nSignif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1")

    def plot(self, plots: Literal["scatter", "box", "density", "bar"] = "scatter", nobs: int = None) -> None:
        """
        Plots the results of the hypothesis test.

        Parameters
        ----------
        plots : str
            The type of plot to create ('scatter', 'box', 'density', 'bar').
        nobs : int, optional
            The number of observations to plot (default is None in which case all available data points will be used).
        """
        if plots == "scatter":
            if (
                nobs is not None
                and nobs < self.data.shape[0]
                and nobs != np.inf
                and nobs != -1
            ):
                data = self.data.copy().sample(nobs)
            else:
                data = self.data.copy()

            sns.swarmplot(data=data, x=self.var1, y=self.var2, alpha=0.5)

            # Get the unique categories and their indices
            categories = data[self.var1].cat.categories
            category_indices = {category: i for i, category in enumerate(categories)}

            category_means = data.groupby(self.var1, observed=False)[self.var2].mean()

            # Add a horizontal line for each category at the mean of the value for that category
            for category, mean in category_means.items():
                plt.hlines(
                    y=mean,
                    xmin=category_indices[category] - 0.3,
                    xmax=category_indices[category] + 0.3,
                    color="blue",
                    linestyle="--",
                    linewidth=2,
                    zorder=2,
                )

        elif plots == "box":
            sns.boxplot(data=self.data, x=self.var1, y=self.var2)
        elif plots == "density":
            sns.kdeplot(data=self.data, x=self.var2, hue=self.var1)
        elif plots == "bar":
            sns.barplot(
                data=self.data,
                y=self.var2,
                x=self.var1,
                # yerr=self.descriptive_stats["se"], # shapes don't align for some reason
            )
        else:
            print("Invalid plot type")



from typing import Optional, Literal
import pandas as pd
import polars as pl
import matplotlib.pyplot as plt

# from sklearn.model_selection import GridSearchCV
# from sklearn.model_selection import StratifiedKFold
from sklearn.neural_network import MLPClassifier, MLPRegressor
from sklearn.inspection import PartialDependenceDisplay as pdp
from pyrsm.utils import ifelse, check_dataframe, setdiff
from pyrsm.model.model import (
    sim_prediction,
    convert_binary,
    evalreg,
    convert_to_list,
    reg_dashboard,
)
from pyrsm.model.perf import auc
from pyrsm.stats import scale_df
from .visualize import pred_plot_sk, vimp_plot_sk

# update docstrings using https://chatgpt.com/share/e/95faf46d-7f74-4ab7-b24b-fac0d22e6dec


class mlp:
    """
    Initialize a Multi-layer Perceptron (NN) model.

    Parameters
    ----------
    data : pd.DataFrame or pl.DataFrame or dict of str to pd.DataFrame or pl.DataFrame
        The dataset to be used. If a dictionary is provided, the key will be used as the dataset name.
    rvar : str, optional
        The name of the column to be used as the response variable.
    lev : str, optional
        The level in the response variable to be modeled.
    evar : list of str, optional
        The names of the columns to be used as explanatory (target) variables.
    hidden_layer_sizes : tuple, default=(5,)
        The number of neurons in the hidden layers. For example, (5,) for 5 neurons in 1 hidden layer, (5, 5) for 5 neurons in 2 hidden layers, etc.
    activation : {'identity', 'logistic', 'tanh', 'relu'}, default='tanh'
        Activation function to transform the data at each node in the hidden layer.
    solver : {'lbfgs', 'sgd', 'adam'}, default='lbfgs'
        The solver for weight optimization. Note that 'adam' also uses stochastic gradient descent.
    alpha : float, default=0.0001
        L2 penalty (regularization term) parameter.
    batch_size : float or str, default='auto'
        Size of minibatches for stochastic optimizers (i.e., 'sgd' or 'adam'). If 'auto', batch_size=min(200, n_samples).
    learning_rate_init : float, default=0.001
        Initial learning rate used. It controls the step-size in updating the weights. Only used when solver='sgd' or 'adam'.
    max_iter : int, default=1_000_000
        Maximum number of iterations. The solver iterates until convergence (determined by 'tol') or this number of iterations.
    random_state : int, default=1234
        Seed of the pseudo-random number generator to use for shuffling the data.
    mod_type : {'regression', 'classification'}, default='classification'
        Type of model to fit, either regression or classification.
    **kwargs : dict
        Additional arguments to be passed to the sklearn's Multi-layer Perceptron functions.

    Examples
    --------
    >>> model = mlp(data=df, rvar='target', evar=['feature1', 'feature2'])

    """

    def __init__(
        self,
        data: pd.DataFrame | pl.DataFrame | dict[str, pd.DataFrame | pl.DataFrame],
        rvar: Optional[str] = None,
        lev: Optional[str] = None,
        evar: Optional[list[str]] = None,
        hidden_layer_sizes: tuple = (5,),
        activation: Literal["identity", "logistic", "tanh", "relu"] = "tanh",
        solver: Literal["lbfgs", "sgd", "adam"] = "lbfgs",
        alpha: float = 0.0001,
        batch_size: float | str = "auto",
        learning_rate_init: float = 0.001,
        max_iter: int = 1_000_000,
        random_state: int = 1234,
        mod_type: Literal["regression", "classification"] = "classification",
        **kwargs,
    ) -> None:
        if isinstance(data, dict):
            self.name = list(data.keys())[0]
            self.data = data[self.name]
        else:
            self.data = data
            self.name = "Not provided"
        self.data = check_dataframe(self.data)
        self.rvar = rvar
        self.lev = lev
        self.evar = convert_to_list(evar)
        self.mod_type = mod_type
        self.hidden_layer_sizes = hidden_layer_sizes
        self.activation = activation
        self.solver = solver
        self.alpha = alpha
        self.batch_size = batch_size
        self.learning_rate_init = learning_rate_init
        self.max_iter = max_iter
        self.random_state = random_state
        self.kwargs = kwargs

        if self.mod_type == "classification":
            if self.lev is not None and self.rvar is not None:
                self.data = convert_binary(self.data, self.rvar, self.lev)

            self.mlp = MLPClassifier(
                hidden_layer_sizes=self.hidden_layer_sizes,
                activation=self.activation,
                solver=self.solver,
                alpha=self.alpha,
                batch_size=self.batch_size,
                learning_rate_init=self.learning_rate_init,
                max_iter=self.max_iter,
                random_state=self.random_state,
                **kwargs,
            )
        else:
            self.mlp = MLPRegressor(
                hidden_layer_sizes=self.hidden_layer_sizes,
                activation=self.activation,
                solver=self.solver,
                alpha=self.alpha,
                batch_size=self.batch_size,
                learning_rate_init=self.learning_rate_init,
                max_iter=self.max_iter,
                random_state=self.random_state,
                **kwargs,
            )
        self.data_std, self.means, self.stds = scale_df(self.data[[rvar] + self.evar], sf=1, stats=True)
        # use drop_first=True for one-hot encoding because NN models include a bias term
        self.data_onehot = pd.get_dummies(self.data_std[self.evar], drop_first=True)
        self.n_features = [len(evar), self.data_onehot.shape[1]]

        self.fitted = self.mlp.fit(self.data_onehot, self.data_std[self.rvar])
        self.n_weights = sum(weight_matrix.size for weight_matrix in self.fitted.coefs_)
        self.nobs = self.data.dropna().shape[0]

    def summary(self, dec=3) -> None:
        """
        Summarize the output from a Multi-layer Perceptron (NN) model.

        Parameters
        ----------
        dec : int, default=3
            Number of decimal places to display in the summary.

        Examples
        --------
        >>> model.summary()
        """
        print("Multi-layer Perceptron (NN)")
        print(f"Data                 : {self.name}")
        print(f"Response variable    : {self.rvar}")
        if self.mod_type == "classification":
            print(f"Level                : {self.lev}")
        print(f"Explanatory variables: {', '.join(self.evar)}")
        print(f"Model type           : {ifelse(self.mod_type == 'classification', 'classification', 'regression')}")
        print(f"Nr. of features      : ({self.n_features[0]}, {self.n_features[1]})")
        print(f"Nr. of weights       : {format(self.n_weights, ',.0f')}")
        print(f"Nr. of observations  : {format(self.nobs, ',.0f')}")
        print(f"Hidden_layer_sizes   : {self.hidden_layer_sizes}")
        print(f"Activation function  : {self.activation}")
        print(f"Solver               : {self.solver}")
        print(f"Alpha                : {self.alpha}")
        print(f"Batch size           : {self.batch_size}")
        print(f"Learning rate        : {self.learning_rate_init}")
        print(f"Maximum iterations   : {self.max_iter}")
        print(f"random_state         : {self.random_state}")
        if self.mod_type == "classification":
            print(
                f"AUC                  : {round(auc(self.data[self.rvar], self.fitted.predict_proba(self.data_onehot)[:, -1]), dec)}"
            )
        else:
            print("Model fit            :")
            print(
                evalreg(
                    pd.DataFrame().assign(
                        rvar=self.data_std[[self.rvar]],
                        prediction=self.fitted.predict(self.data_onehot),
                    ),
                    "rvar",
                    "prediction",
                    dec=dec,
                )
                .T[2:]
                .rename(columns={0: " "})
                .T.to_string()
            )

        if len(self.kwargs) > 0:
            kwargs_list = [f"{k}={v}" for k, v in self.kwargs.items()]
            print(f"Extra arguments      : {', '.join(kwargs_list)}")

        print("\nRaw data             :")
        print(self.data[self.evar].head().to_string(index=False))

        print("\nEstimation data      :")
        print(self.data_onehot.head().to_string(index=False))

    def predict(self, data=None, cmd=None, data_cmd=None, scale=True, means=None, stds=None) -> pd.DataFrame:
        """
        Predict probabilities or values for the MLP model.

        Parameters
        ----------
        data : pd.DataFrame, optional
            The data to predict. If None, uses the training data.
        cmd : dict, optional
            Command dictionary to simulate predictions.
        data_cmd : dict, optional
            Command dictionary to modify the data before predictions.
        scale : bool, default=True
            Whether to scale the data before prediction.
        means : pd.Series, optional
            Means of the training data features for scaling. Will use the means used during estimation if not provided.
        stds : pd.Series, optional
            Standard deviations of the training data features for scaling. Will use the standard deviations used during estimation if not provided.

        Returns
        -------
        pd.DataFrame
            DataFrame with predictions.

        Examples
        --------
        >>> predictions = model.predict(new_data)
        """
        if data is None:
            data = self.data.loc[:, self.evar].copy()
        else:
            data = data.loc[:, self.evar].copy()

        if data_cmd is not None and data_cmd != "":
            for k, v in data_cmd.items():
                data[k] = v
        elif cmd is not None and cmd != "":
            cmd = {k: ifelse(isinstance(v, str), [v], v) for k, v in cmd.items()}
            data = sim_prediction(data=data, vary=cmd)

        if scale or (means is not None and stds is not None):
            if means is not None and stds is not None:
                data_std = scale_df(data, sf=1, means=means, stds=stds)
            else:
                # scaling the full dataset by the means used during estimation
                data_std = scale_df(data, sf=1, means=self.means, stds=self.stds)

            data_onehot = pd.get_dummies(data_std, drop_first=False)
        else:
            data_onehot = pd.get_dummies(data, drop_first=False)

        # adding back levels for categorical variables is they were removed
        if data_onehot.shape[1] != self.data_onehot.shape[1]:
            for k in setdiff(self.data_onehot.columns, data_onehot.columns):
                data_onehot[k] = False
            data_onehot = data_onehot[self.data_onehot.columns]

        if self.mod_type == "classification":
            data["prediction"] = self.fitted.predict_proba(data_onehot)[:, -1]
        else:
            print(data_onehot.head())
            data["prediction"] = self.fitted.predict(data_onehot) * self.stds[self.rvar] + self.means[self.rvar]

        return data

    def plot(
        self,
        plots: Literal["pred", "pdp", "vimp", "dashboard"] = "pred",
        data=None,
        incl=None,
        excl=None,
        incl_int=[],
        nobs: int = 1000,
        fix=True,
        hline=False,
        nnv=40,
        minq=0.025,
        maxq=0.975,
        figsize=None,
        ax=None,
        ret=None,
    ) -> None:
        """
        Generate plots for the Multi-layer Perceptron model.

        Parameters
        ----------
        plots : {'pred', 'pdp', 'vimp', 'dashboard'}, default='pred'
            Type of plot to generate. Options are 'pred' for prediction plot, 'pdp' for partial dependence plot, 'vimp' for variable importance plot which uses permutation importance, and 'dashboard' for a regression dashboard.
        data : pd.DataFrame, optional
            Data to use for the plots. If None, uses the training data.
        incl : list of str, optional
            Variables to include in the plots.
        excl : list of str, optional
            Variables to exclude from the plots.
        incl_int : list, optional
            Interactions to include in the plots.
        nobs : int, default=1000
            Number of observations to include in the scatter plots for the dashboard plot.
        fix : bool, default=True
            Whether to fix the scale of the plots based on the maximum impact range of the include explanatory variables.
        hline : bool, default=False
            Whether to include a horizontal line at the mean response rate in the plots.
        nnv : int, default=40
            Number of values to use for the prediction plot.
        minq : float, default=0.025
            Minimum quantile for the prediction plot.
        maxq : float, default=0.975
            Maximum quantile for the prediction plot.
        figsize : tuple, optional
            Figure size for the plots.
        ax : plt.Axes, optional
            Axes object to plot on.
        ret : bool, optional
            Whether to return the the variable (permutation) importance scores.

        Examples
        --------
        >>> model.plot(plots='pdp')
        >>> model.plot(plots='vimp', data=new_data)
        """
        rvar = self.rvar
        if "pred" in plots:
            if data is None:
                data_dct = {
                    "data": self.data[self.evar + [rvar]],
                    "means": self.means,
                    "stds": self.stds,
                }
            else:
                if self.rvar in data.columns:
                    vars = self.evar + [rvar]
                    if self.mod_type == "classification":
                        data = convert_binary(data, rvar, self.lev)
                else:
                    vars = self.evar
                    rvar = None
                data_dct = {
                    "data": data[vars],
                    "means": self.means,
                    "stds": self.stds,
                }

            pred_plot_sk(
                self.fitted,
                data=data_dct,
                rvar=rvar,
                incl=incl,
                excl=ifelse(excl is None, [], excl),
                incl_int=incl_int,
                fix=fix,
                hline=hline,
                nnv=nnv,
                minq=minq,
                maxq=maxq,
            )

        if "pdp" in plots:
            if figsize is None:
                figsize = (8, len(self.data_onehot.columns) * 2)
            fig, ax = plt.subplots(figsize=figsize)
            ax.set_title("Partial Dependence Plots")
            fig = pdp.from_estimator(self.fitted, self.data_onehot, self.data_onehot.columns, ax=ax, n_cols=2)

        if "vimp" in plots:
            return_vimp = vimp_plot_sk(
                self.fitted,
                self.data_onehot,
                self.data[self.rvar],
                rep=5,
                ax=ax,
                ret=True,
            )

            if ret is not None:
                return return_vimp

        if "dashboard" in plots and self.mod_type == "regression":
            model = self.fitted
            model.fittedvalues = self.predict()["prediction"]
            model.resid = self.data[self.rvar] - model.fittedvalues
            model.model = pd.DataFrame({"endog": self.data[self.rvar]})
            reg_dashboard(model, nobs=nobs)



import platform
import black
import inspect
from shiny import req
from htmltools import tags, div, css
from itertools import combinations
from shiny import render, ui, reactive
from faicons import icon_svg
import pandas as pd
import polars as pl
from pyrsm.utils import ifelse, check_dataframe
from pyrsm.example_data import load_data


def set_host(host):
    if host == "":
        if platform.system() == "Linux":
            return "0.0.0.0"
        else:
            return "localhost"
    else:
        return host


def get_dfs(pkg=None, name=None, polars=False):
    obj = pd.DataFrame if not polars else pl.DataFrame
    data_dct = {k: v for k, v in globals().items() if isinstance(v, obj)}
    descriptions_dct = {
        k: globals()[f"{k}_description"]
        for k in data_dct
        if f"{k}_description" in globals()
    }
    if len(data_dct) == 0 and name is not None:
        data, description = load_data(pkg=pkg, name=name, polars=polars)
        data_dct = {name: data}
        descriptions_dct = {name: description}
    elif len(data_dct) == 0 and pkg is not None:
        data_dct, descriptions_dct = load_data(pkg=pkg)

    return ifelse(len(data_dct) == 0, None, data_dct), ifelse(
        len(descriptions_dct) == 0, None, descriptions_dct
    )


def message():
    print(
        "Pyrsm and Radiant are open source tools and free to use. If you\nare a student or instructor using pyrsm or Radiant for a class,\nas a favor to the developers, please send an email to\n<radiant@rady.ucsd.edu> with the name of the school and class.\nIf you are using Radiant in your company, as a favor to the\ndeveloper, please share the name of your company and what types\nof activites you are supporting with the tool."
    )


def head_content():
    """
    Return the head content for the shiny app
    """
    return ui.head_content(
        # from https://github.com/rstudio/py-shiny/issues/491#issuecomment-1579138681
        ui.tags.link(
            href="//cdnjs.cloudflare.com/ajax/libs/highlight.js/11.8.0/styles/agate.min.css",
            rel="stylesheet",
        ),
        ui.tags.link(rel="shortcut icon", href="/www/imgs/icon.png"),
        ui.tags.link(href="/www/style.css", rel="stylesheet"),
        ui.tags.script(src="/www/js/returnTextAreaBinding.js"),
        ui.tags.script(src="/www/js/radiantUI.js"),
        ui.tags.script(src="/www/js/screenshot.js"),
        ui.tags.script(
            src="//cdnjs.cloudflare.com/ajax/libs/highlight.js/11.8.0/highlight.min.js",
        ),
        ui.tags.script(
            src="//html2canvas.hertzen.com/dist/html2canvas.min.js",
        ),
    )


def init(self, datasets, descriptions=None, state=None, code=True, navbar=None):
    """
    Initialize key 'self' values to be used in an app class
    """
    datasets = ifelse(isinstance(datasets, dict), datasets, {"dataset": datasets})
    self.datasets = {k: check_dataframe(d) for k, d in datasets.items()}
    if descriptions is None:
        self.descriptions = {
            k: "## No data description provided" for k in self.datasets.keys()
        }
    else:
        self.descriptions = ifelse(
            isinstance(descriptions, dict),
            descriptions,
            {"description": descriptions},
        )
    self.dataset_list = list(datasets.keys())
    self.code = code  # keep code windows open or closed by default
    self.stop_code = ""
    if state is None:
        self.state = {"__pending_changes__": False}
    else:
        self.state = state
        self.state["__pending_changes__"] = False
    if navbar is None:
        self.navbar = ()
    else:
        self.navbar = navbar


def escape_quotes(cmd):
    return cmd.replace('"', '\\"').replace("'", "\\'")


def quote(v, k, ignore=["data"]):
    if isinstance(v, str) and k not in ignore and not (v[0] + v[-1] == "{}"):
        v = escape_quotes(v)
        return f'"{v}"'
    else:
        return v


def copy_icon(cmd, id):
    cmd = escape_quotes(cmd).replace("\n", "\\n")
    print(cmd)
    return (
        ui.input_action_link(
            id,
            None,
            icon=icon_svg("copy", width="1.5em", height="1.5em"),
            title="Copy to clipboard",
            onclick=f"""copyToClipboard("{cmd}");""",
        ),
    )


def copy_reset(input, session, id="copy"):
    @reactive.Effect
    @reactive.event(input[id], ignore_none=True)
    async def copy_success():
        ui.update_action_link(
            id,
            icon=icon_svg("check", fill="green", width="1.5em", height="1.5em"),
        )
        # await session.send_custom_message("reset_copy", "")

    @reactive.Effect
    @reactive.event(input.copy_reset, ignore_none=True)
    async def copy_reset():
        req(input[id])
        ui.update_action_link(
            id,
            icon=icon_svg("copy", width="1.5em", height="1.5em"),
        )


def code_formatter(code, self, input, session, id="copy"):
    """
    Format python code using black
    """
    cmd = self.stop_code = black.format_str(code, mode=black.Mode())
    print(cmd)
    copy_reset(input, session, id=id)
    return ui.TagList(
        ui.HTML(
            f"<details {ifelse(self.code, 'open', '')}><summary>View generated Python code</summary>"
        ),
        copy_icon(cmd, id=id),
        ui.markdown(f"""\n```python\n{cmd.rstrip()}\n```"""),
        ui.tags.script(
            "hljs.highlightAll(); hljs.configure({ignoreUnescapedHTML: true});"
        ),
        ui.HTML("</details>"),
        ui.br(),
    )


def drop_default_args(args, func, ignore=["data"]):
    """
    Take a dictionary of arguments for a function and compare
    to the default arguments for that function
    Return a comma separate string of key-value pairs for
    non-default values where the value is quoted if it is a string
    """
    sig = inspect.signature(func).parameters
    keep = {k: args[k] for k, v in sig.items() if k in args and args[k] != v.default}

    return ", ".join(f"{k}={quote(v, k, ignore)}" for k, v in keep.items())


def make_side_by_side(a, b):
    """Put two inputs side-by-side in the UI"""
    return ui.tags.table(
        ui.tags.td(a, width="50%"), ui.tags.td(b, width="50%"), width="100%"
    )


def get_data(self, input):
    """
    Set up data access for the app
    """
    data_name = input.datasets()
    data = self.datasets[data_name]
    description = self.descriptions[data_name]
    code = (
        f"import pyrsm as rsm\n# {data_name} = pd.read_parquet('{data_name}.parquet')"
    )
    if input.show_filter():
        code_sf = ""
        if not is_empty(input.data_filter()):
            code_sf += f""".query("{escape_quotes(input.data_filter())}")"""
        if not is_empty(input.data_sort()):
            code_sf += f""".sort_values({input.data_sort()})"""
        if not is_empty(input.data_slice()):
            code_sf += f""".iloc[{input.data_slice()}, :]"""

        if not is_empty(code_sf):
            data = eval(f"""data{code_sf}""")
            code = f"""{code}\n{data_name} = {data_name}{code_sf}"""

    types = {c: [data[c].dtype, data[c].nunique()] for c in data.columns}
    isNum = {
        c: f"{c} ({t[0].name})"
        for c, t in types.items()
        if pd.api.types.is_numeric_dtype(t[0])
    }
    isBin = {c: f"{c} ({t[0].name})" for c, t in types.items() if t[1] == 2}
    isCat = {
        c: f"{c} ({t[0].name})"
        for c, t in types.items()
        if c in isBin or pd.api.types.is_categorical_dtype(t[0]) or t[1] < 10
    }
    var_types = {
        "all": {c: f"{c} ({t[0].name})" for c, t in types.items()},
        "isNum": isNum,
        "isBin": isBin,
        "isCat": isCat,
    }

    return {
        "data": data,
        "data_name": data_name,
        "description": description,
        "var_types": var_types,
        "code": code,
    }


def make_data_elements(self, input, output, session):
    @output(id="show_data_code")
    @render.ui
    def show_data_code():
        return code_formatter(
            get_data(self, input)["code"], self, input, session, id="copy_data"
        )

    @output(id="show_data")
    @render.data_frame
    def show_data():
        data = get_data(self, input)["data"]
        summary = "Viewing rows {start} through {end} of {total}"
        if data.shape[0] > 100_000:
            data = data[:100_000]
            summary += " (100K rows shown)"

        return render.DataTable(data, summary=summary)

    @output(id="show_description")
    @render.ui
    def show_description():
        return ui.markdown(get_data(self, input)["description"])

    @reactive.Calc
    def reactive_get_data():
        return get_data(self, input)

    return reactive_get_data


def check_input_value(k, i):
    """
    The '... in input' approach fails with some regulariy (see app_state.py).
    see https://discord.com/channels/1109483223987277844/1127817202804985917/1129530385429176371
    Also tuples need to be converted to lists to be picked up on refresh. Not
    clear why that is needed but it is
    """
    try:
        # print(f"Working on {k}")
        value = i[k]()
        if isinstance(value, tuple):
            return list(value)
        else:
            return value
    except Exception as err:
        # print(f"Extracting value for {k} failed")
        print(err)  # no error message printed
        return None


def dct_update(self, input):
    input_keys = [k for k in input.__dict__["_map"].keys() if k[0] != "."]
    self.state.update({k: check_input_value(k, input) for k in input_keys})
    self.state["__pending_changes__"] = False


# def update_state():
#     with reactive.isolate():
#         avoid = ["copy_reset"]  # issue because async?
#         input_keys = [
#             k
#             for k in input.__dict__["_map"].keys()
#             if k[0] != "." and k not in avoid
#         ]
#         # print(type(input.copy_reset())) # can't even run this without an error
#         self.state.update({k: input[k]() for k in input_keys})
#         self.state.update(
#             {k: list(v) for k, v in self.state.items() if isinstance(v, tuple)}
#         )


def input_return_text_area(id, label, value="", rows=1, placeholder=""):
    classes = ["form-control", "returnTextArea"]
    area = tags.textarea(
        value,
        id=id,
        class_=" ".join(classes),
        style=css(width="100%", height=None, resize="vertical"),
        placeholder=placeholder,
        rows=rows,
        cols=None,
        autocomplete=False,
        spellcheck=False,
    )

    def shiny_input_label(id, label=None):
        cls = "control-label" + ("" if label else " shiny-label-null")
        return tags.label(label, class_=cls, id=id + "-label", for_=id)

    return div(
        shiny_input_label(id, label),
        area,
        None,
        class_="form-group shiny-input-container",
        style=css(width="100%"),
    )


def is_empty(x):
    return x is None or all(c.isspace() for c in x)


def qterms(vars, nway=2):
    return [f"I({v} ** {p})" for p in range(2, nway + 1) for v in vars]


def iterms(vars, nway=2, sep=":"):
    cvars = list(combinations(vars, 2))
    if nway > 2:
        cvars += list(combinations(vars, nway))
    return [f"{sep}".join(c) for c in cvars]


def ui_view(self):
    return ui.panel_well(
        ui.input_select(
            "datasets",
            "Datasets:",
            self.dataset_list,
            selected=self.state.get("datasets", None),
        ),
        ui.input_checkbox(
            "show_filter", "Show data filter", value=self.state.get("show_filter", True)
        ),
        ui.panel_conditional(
            "input.show_filter == true",
            # ui.input_radio_buttons(
            #     "data_language",
            #     "Data language",
            #     ["Pandas", "Polars", "SQL"],
            #     inline=True,
            # ),
            input_return_text_area(
                "data_filter",
                "Data Filter:",
                rows=2,
                value=self.state.get("data_filter", ""),
                placeholder="Provide a filter (e.g., price > 5000) and press return",
            ),
            input_return_text_area(
                "data_sort",
                "Data sort:",
                rows=2,
                value=self.state.get("data_sort", ""),
                placeholder="Sort (e.g., ['color', 'price'], ascending=[True, False])) and press return",
            ),
            input_return_text_area(
                "data_slice",
                "Data slice (rows):",
                rows=1,
                value=self.state.get("data_slice", ""),
                placeholder="e.g., 0:50 and press return",
            ),
        ),
    )


def ui_data(self):
    return ui.panel_conditional("input.tabs == 'Data'", ui_view(self))


def ui_summary(*args):
    return ui.panel_conditional(
        "input.tabs == 'Summary'",
        ui.panel_well(
            ui.input_action_button(
                "run",
                "Estimate model",
                icon=icon_svg("play"),
                class_="btn-success",
                width="100%",
            ),
            ui.output_ui("ui_rvar"),
            ui.panel_conditional(
                "input.mod_type === 'classification' | typeof input.mod_type === 'undefined'",
                ui.output_ui("ui_lev"),
            ),
            ui.output_ui("ui_evar"),
            *args,
        ),
    )


def ui_plot(self, choices, *args):
    return ui.panel_conditional(
        "input.tabs == 'Plot'",
        ui.panel_well(
            ui.input_select(
                id="plots",
                label="Plots",
                selected=self.state.get("plots", None),
                choices=choices,
            ),
            *args,
        ),
    )


def ui_data_main():
    return ui.div(
        ui.output_ui("show_data_code"),
        ui.output_data_frame("show_data"),
        ui.output_ui("show_description"),
    )


def ui_main_basics(height="500px", width="700px"):
    return ui.navset_card_tab(
        ui.nav_panel(
            "Data",
            ui_data_main(),
        ),
        ui.nav_panel(
            "Summary",
            ui.output_ui("show_summary_code"),
            ui.output_text_verbatim("summary"),
        ),
        ui.nav_panel(
            "Plot",
            ui.output_ui("show_plot_code"),
            ui.output_ui("plot_container"),
        ),
        id="tabs",
    )


def ui_main_model():
    return ui.navset_card_tab(
        ui.nav_panel(
            "Data",
            ui_data_main(),
        ),
        ui.nav_panel(
            "Summary",
            ui.output_ui("show_estimation_code"),
            ui.output_ui("show_summary_code"),
            ui.output_text_verbatim("summary"),
        ),
        ui.nav_panel(
            "Predict",
            ui.output_ui("show_predict_code"),
            ui.output_data_frame("predict"),
        ),
        ui.nav_panel(
            "Plot",
            ui.output_ui("show_plot_code"),
            # ui.output_plot("plot", height="800px", width="700px"),
            ui.output_ui("plot_container"),
        ),
        id="tabs",
    )


def radiant_navbar():
    return (
        ui.nav_control(
            ui.input_action_link("data", "Data", onclick='window.location.href = "/";')
        ),
        ui.nav_menu(
            "Basics",
            "Probability",
            ui.nav_control(
                ui.input_action_link(
                    "pc",
                    "Probability Calculator",
                    onclick='window.location.href = "/basics/prob-calc/";',
                ),
            ),
            "-----",
            "Means",
            ui.nav_control(
                ui.input_action_link(
                    "sm",
                    "Single mean",
                    onclick='window.location.href = "/basics/single-mean/";',
                ),
                ui.input_action_link(
                    "cm",
                    "Compare means",
                    onclick='window.location.href = "/basics/compare-means/";',
                ),
            ),
            "----",
            "Proportions",
            ui.nav_control(
                ui.input_action_link(
                    "sp",
                    "Single proportion",
                    onclick='window.location.href = "/basics/single-prop/";',
                ),
                ui.input_action_link(
                    "cp",
                    "Compare proportions",
                    onclick='window.location.href = "/basics/compare-props/";',
                ),
            ),
            "----",
            "Tables",
            ui.nav_control(
                ui.input_action_link(
                    "gt",
                    "Goodness of fit",
                    onclick='window.location.href = "/basics/goodness/";',
                ),
                ui.input_action_link(
                    "ct",
                    "Cross-tabs",
                    onclick='window.location.href = "/basics/cross-tabs/";',
                ),
                ui.input_action_link(
                    "cr",
                    "Correlation",
                    onclick='window.location.href = "/basics/correlation/";',
                ),
            ),
        ),
        ui.nav_menu(
            "Models",
            "Estimate",
            ui.nav_control(
                ui.input_action_link(
                    "reg",
                    "Linear Regression (OLS)",
                    onclick='window.location.href = "/models/regress/";',
                ),
                ui.input_action_link(
                    "lr",
                    "Logistic Regression (GLM)",
                    onclick='window.location.href = "/models/logistic/";',
                ),
            ),
        ),
    )


def ui_stop():
    return (
        # ui.nav_control(
        #     ui.input_action_link(
        #         "screenshot",
        #         "Screenshot",
        #         icon=icon_svg("camera"),
        #         onclick="generate_screenshot();",
        #     ),
        # ),
        ui.nav_control(
            ui.input_action_link(
                "stop", "Stop", icon=icon_svg("stop"), onclick="window.close();"
            ),
        ),
    )


def ui_help(link, example):
    return (
        ui.nav_menu(
            "Help",
            ui.nav_control(
                ui.a(
                    icon_svg("question"),
                    example,
                    href=link,
                    target="_blank",
                ),
            ),
            ui.nav_control(
                ui.a(
                    icon_svg("github"),
                    "Radiant-for-python source code",
                    href="https://github.com/vnijs/pyrsm/tree/main/pyrsm/radiant",
                    target="_blank",
                ),
            ),
            ui.nav_control(
                ui.a(
                    icon_svg("github"),
                    "Pyrsm source code",
                    href="https://github.com/vnijs/pyrsm/tree/main",
                    target="_blank",
                ),
            ),
            ui.nav_control(
                ui.a(
                    icon_svg("docker"),
                    "Rady MSBA docker container",
                    href="https://github.com/radiant-rstats/docker",
                    target="_blank",
                ),
            ),
            align="right",
        ),
    )


def reestimate(input):
    @reactive.Effect
    def run_refresh():
        def update():
            with reactive.isolate():
                if input.run() > 0:  # only update if run button was pressed
                    ui.update_action_button(
                        "run",
                        label="Re-estimate model",
                        icon=icon_svg("rotate"),
                    )

        if not is_empty(input.evar()) and not is_empty(input.rvar()):
            update()

        if not is_empty(input.data_filter()) or not is_empty(input.data_slice()):
            update()

        # not clear why this needs to be separate from the above
        if not is_empty(input.interactions()):
            update()

        if not is_empty(input.hidden_layer_sizes()):
            update()

        if not is_empty(input.alpha()):
            update()

    @reactive.Effect
    @reactive.event(input.run, ignore_none=True)
    def run_done():
        ui.update_action_button(
            "run",
            label="Estimate model",
            icon=icon_svg("play"),
        )

    return run_refresh, run_done  # , copy_success, copy_reset


# radiant_screenshot_modal <- function(report_on = "") {
#   add_button <- function() {
#     if (is.empty(report_on)) {
#       ""
#     } else {
#       actionButton(report_on, "Report", icon = icon("edit", verify_fa = FALSE), class = "btn-success")
#     }
#   }
#   showModal(
#     modalDialog(
#       title = "Radiant screenshot",
#       span(shiny::tags$div(id = "screenshot_preview")),
#       span(HTML("</br>To include a screenshot in a report first save it to disk by clicking on the <em>Save</em> button. Then click the <em>Report</em> button to insert a reference to the screenshot into <em>Report > Rmd</em>.")),
#       footer = tagList(
#         tags$table(
#           tags$td(download_button("screenshot_save", "Save", ic = "download")),
#           tags$td(add_button()),
#           tags$td(modalButton("Cancel")),
#           align = "right"
#         )
#       ),
#       size = "l",
#       easyClose = TRUE
#     )
#   )
# }

# observeEvent(input$screenshot_link, {
#   radiant_screenshot_modal()
# })

# render_screenshot <- function() {
#   plt <- sub("data:.+base64,", "", input$img_src)
#   png::readPNG(base64enc::base64decode(what = plt))
# }

# download_handler_screenshot <- function(path, plot, ...) {
#   plot <- try(plot(), silent = TRUE)
#   if (inherits(plot, "try-error") || is.character(plot) || is.null(plot)) {
#     plot <- ggplot() +
#       labs(title = "Plot not available")
#     png(file = path, width = 500, height = 100, res = 96)
#     print(plot)
#     dev.off()
#   } else {
#     ppath <- parse_path(path, pdir = getOption("radiant.launch_dir", find_home()), mess = FALSE)
#     # r_info[["latest_screenshot"]] <- glue("![]({ppath$rpath})")
#     # r_info[["latest_screenshot"]] <- glue("<details>\n<summary>Click to show screenshot</summary>\n<img src='{ppath$rpath}' alt='Radiant screenshot'>\n</details>")
#     r_info[["latest_screenshot"]] <- glue("\n<details>\n<summary>Click to show screenshot with Radiant settings to generate output shown below</summary>\n\n![]({ppath$rpath})\n</details></br>\n")
#     png::writePNG(plot, path, dpi = 144)
#   }
# }

# download_handler(
#   id = "screenshot_save",
#   fun = download_handler_screenshot,
#   fn = function() paste0(r_info[["radiant_tab_name"]], "-screenshot"),
#   type = "png",
#   caption = "Save radiant screenshot",
#   plot = render_screenshot,
#   btn = "button",
#   label = "Save",
#   class = "btn-primary",
#   onclick = "get_img_src();"
# )



import os
from importlib import import_module
import pandas as pd
import polars as pl


def load_data(pkg=None, name=None, dct=None, polars=False):
    """
    Load example data included in the pyrsm package

    Parameters
    ----------
    pkg : str
        One of "data", "design", "basics", "model", or "multivariate"
        These string coincide with the dropdown menus at https://radiant-rstats.github.io/docs/
        If None, datasets from all packages will be returned
    name : str
        Name of the dataset you want to load (e.g., "diamonds").
        If None, all datasets will be returned
    dct : dct
        Dictionary to add datasets to. For example, using globals() will add
        DataFrames to the global environment in python. If None, a dictionary
        with datasets will be returned
    polars : bool
        Should the returned DataFrame be in Polars format? If False, a Pandas
        DataFrame will be returned

    Examples
    --------

    import pyrsm as rsm
    rsm.load_data(pkg="basics", name="demand_uk", dct=globals())
    data, description = rsm.load_data(pkg="basics", name="demand_uk")
    """

    base = "pyrsm.data"
    base_path = import_module(base).__path__[0]

    def load_data(file_path):
        if polars:
            return pl.read_parquet(file_path)
        else:
            return pd.read_parquet(file_path)

    def load_description(file_path):
        # read a text file into a string
        with open(f"{file_path}_description.md", "r") as file:
            descr = file.read()
        return descr

    def mkdct(spkg):
        data = {}
        description = {}
        for sp in spkg:
            package_path = os.path.join(base_path, sp)
            for file_name in os.listdir(package_path):
                if (name is None and file_name.endswith(".parquet")) or (
                    name is not None and file_name == f"{name}.parquet"
                ):
                    file_path = os.path.join(package_path, file_name)
                    key = file_name.replace(".parquet", "")
                    data[key] = load_data(file_path)
                    description[key] = load_description(
                        file_path.replace(".parquet", "")
                    )
        return data, description

    if pkg is None and name is None:
        data, description = mkdct(
            [
                d
                for d in os.listdir(base_path)
                if not d.startswith("__") and not d.startswith(".")
            ]
        )
    elif pkg is not None and name is None:
        data, description = mkdct([pkg])
    elif pkg is None and name is not None:
        data, description = mkdct(
            [
                d
                for d in os.listdir(base_path)
                if not d.startswith("__") and not d.startswith(".")
            ]
        )
        if dct is None:
            data = data[name]
            description = description[name]
    elif pkg is not None and name is not None:
        data = load_data(os.path.join(base_path, pkg, f"{name}.parquet"))
        description = load_description(os.path.join(base_path, pkg, name))
        if dct is not None:
            data = {name: data}
            description = {name: description}
    else:
        data = {}
        description = {}

    if dct is None:
        return data, description
    else:
        for key, val in data.items():
            dct[key] = val
        for key, val in description.items():
            dct[f"{key}_description"] = val


if __name__ == "__main__":
    data_dct, description_dct = load_data()
    data_dct, description_dct = load_data(pkg="basics")
    demand_uk, demand_uk_description = load_data(pkg="basics", name="demand_uk")
    data_dct, description_dct = load_data(pkg="design")
    diamonds, diamonds_description = load_data(pkg="data", name="diamonds")
    data, description = load_data(pkg="model")
    load_data(pkg="model", name="catalog", dct=globals())
    rndnames, rndnames_description = load_data(name="rndnames")
